# Copyright 2023 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import struct

from chirp import (
    bitwise,
    chirp_common,
    directory,
    errors,
    memmap,
    util,
)
from chirp.settings import (
    RadioSetting,
    RadioSettingGroup,
    RadioSettings,
    RadioSettingValueBoolean,
    RadioSettingValueInteger,
    RadioSettingValueList,
    RadioSettingValueString,
)

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct {
  lbcd rxfreq[4];     // 0-3
  lbcd txfreq[4];     // 4-7
  ul16 rxtone;        // 8-9
  ul16 txtone;        // A-B
  u8 unknown1:4,      // C
     scode:4;         //     Signaling
  u8 unknown2:6,      // D
     pttid:2;         //     PTT-ID
  u8 unknown3:6,      // E
     txpower:2;       //     Power Level  0 = H, 1 = L, 2 = M
  u8 unknown4:1,      // F
     narrow:1,        //     Bandwidth  0 = Wide, 1 = Narrow
     encrypt:2,       //     Encrypt
     bcl:1,           //     BCL
     scan:1,          //     Scan  0 = Skip, 1 = Scan
     unknown5:1,      //
     learning:1;      //     Learning
  lbcd code[3];       // 0-2 Code
  u8 unknown6;        // 3
  char name[12];      // 4-F 12-character Alpha Tag
} memory[%d];

#seekto 0x9000;
struct {
  u8 unused:4,        // 9000
     sql:4;           //      Squelch Level
  u8 unused_9001:6,   // 9001
     save:2;          //      Battery Save
  u8 unused_9002:4,   // 9002
     vox:4;           //      VOX Level
  u8 unused_9003:4,   // 9003
     abr:4;           //      Auto BackLight
  u8 unused_9004:7,   // 9004
     tdr:1;           //      TDR
  u8 unused_9005:5,   // 9005
     tot:3;           //      Time-out Timer
  u8 unused_9006:7,   // 9006
     beep:1;          //      Beep
  u8 unused_9007:7,   // 9007
     voice:1;         //      Voice Prompt
  u8 unused_9008:7,   // 9008
     language:1;      //      Language
  u8 unused_9009:6,   // 9009
     dtmfst:2;        //      DTMF ST
  u8 unused_900a:6,   // 900A
     screv:2;         //      Scan Mode
  u8 unused_900B:4,   // 900B
     pttid:4;         //      PTT-ID
  u8 unused_900c:5,   // 900C
     pttlt:3;         //      PTT Delay
  u8 unused_900d:6,   // 900D
     mdfa:2;          //      Channel_A Display
  u8 unused_900e:6,   // 900E
     mdfb:2;          //      Channel_B Display
  u8 unknown_900f;    // 900F
  u8 unused_9010:4,   // 9010
     autolk:4;        //      Key Auto Lock
  u8 unused_9011:6,   // 9011
     almode:2;        //      Alarm Mode
  u8 unused_9012:7,   // 9012
     alarmsound:1;    //      Alarm Sound
  u8 unused_9013:6,   // 9013
     dualtx:2;        //      Dual TX
  u8 unused_9014:7,   // 9014
     ste:1;           //      Tail Noise Clear
  u8 unused_9015:4,   // 9015
     rpste:4;         //      RPT Noise Clear
  u8 unused_9016:4,   // 9016
     rptrl:4;         //      RPT Noise Delay
  u8 unused_9017:6,   // 9017
     roger:2;         //      Roger
  u8 unknown_9018;    // 9018
  u8 unused_9019:7,   // 9019
     fmradio:1;       //      FM Radio
  u8 unused_901a1:3,  // 901A
     vfomrb:1,        //      WorkMode B
     unused_901a2:3,  //
     vfomra:1;        //      WorkMode A
  u8 unused_901b:7,   // 901B
     kblock:1;        //      KB Lock
  u8 unused_901c:7,   // 901C
     ponmsg:1;        //      Power On Msg
  u8 unknown_901d;    // 901D
  u8 unused_901e:6,   // 901E
     tone:2;          //      Pilot
  u8 unknown_901f;    // 901F
  u8 unused_9020:4,   // 9020
     voxd:4;          //      VOX Delay
  u8 unused_9021:4,   // 9021
     menuquit:4;      //      Menu Auto Quit
  u8 unused_9022:7,   // 9022
     tailcode:1;      //      Tail Code (RT-470L)
  u8 unknown_9023;    // 9023
  u8 unknown_9024;    // 9024
  u8 unknown_9025;    // 9025
  u8 unknown_9026;    // 9026
  u8 unknown_9027;    // 9027
  u8 unknown_9028;    // 9028
  u8 unused_9029:6,   // 9029
     qtsave:2;        //      QT Save Type
  u8 ani;             // 902A ANI
  u8 skey2_sp;        // 902B Skey2 Short
  u8 skey2_lp;        // 902C Skey2 Long
  u8 skey3_sp;        // 902D Skey3 Short
  u8 topkey_sp;       // 902E Top Key (RT-470L)
  u8 unused_902f:6,   // 902F
     rxendtail:2;     //      RX END TAIL (RT-470)
                      //      TAIL PHASE (A36plus)
  u8 skey3_lp;        // 9030 Skey3 Long (RT-470L)
                      //      RX END TAIL (A36plus)
} settings;

#seekto 0xA006;
struct {
  u8 unknown_A006;    // A006
  u8 unused_a007:5,   // A007
     dtmfon:3;        //      DTMF Speed (on time)
  u8 unused_a008:5,   // A008
     dtmfoff:3;       //      DTMF Speed (off time)
} dtmf;

#seekto 0xA020;
struct {
  u8 code[5];         //      5-character DTMF Encoder Groups
  u8 unused[11];
} pttid[15];

#seekto 0xB000;
struct {
  u8 code[3];         //      3-character ANI Code Groups
} anicodes[60];

#seekto 0xB0C0;
struct {
  char name[10];      //      10-character ANI Code Group Names
  u8 unused[6];
} aninames[%d];

"""

MEM_FORMAT_A36PLUS = """
#seekto 0xB200;
struct {
  char name[10];      //      10-character Custom CH Names (Talkpod A36plus)
  u8 unused[6];
} customnames[30];

"""


CMD_ACK = b"\x06"

DTCS = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

DTMF_CHARS = "0123456789 *#ABCD"

TXPOWER_HIGH = 0x00
TXPOWER_LOW = 0x01
TXPOWER_MID = 0x02

ABR_LIST = ["On", "5 seconds", "10 seconds", "15 seconds", "20 seconds",
            "30 seconds", "1 minute", "2 minutes", "3 minutes"]
ALMODE_LIST = ["Site", "Tone", "Code"]
AUTOLK_LIST = ["Off"] + ABR_LIST[1:4]
DTMFSPEED_LIST = ["50 ms", "100 ms", "200 ms", "300 ms", "500 ms"]
DTMFST_LIST = ["Off", "KeyBoard Side Tone", "ANI Side Tone", "KB ST + ANI ST"]
DUALTX_LIST = ["Off", "A", "B"]
ENCRYPT_LIST = ["Off", "DCP1", "DCP2", "DCP3"]
LANGUAGE_LIST = ["English", "Chinese"]
MDF_LIST = ["Name", "Frequency", "Channel"]
MENUQUIT_LIST = ["%s seconds" % x for x in range(5, 55, 5)] + ["60 seconds"]
OFF1TO9_LIST = ["Off"] + ["%s" % x for x in range(1, 10)]
PONMSG_LIST = ["Logo", "Voltage"]
PTTID_LIST = ["Off", "BOT", "EOT", "Both"]
PTTIDCODE_LIST = ["%s" % x for x in range(1, 16)]
PTTLT_LIST = ["None", "100 ms"] + \
             ["%s ms" % x for x in range(200, 1200, 200)]
QTSAVE_LIST = ["All", "RX", "TX"]
ROGER_LIST = ["OFF", "BEEP", "TONE1200"]
RPSTE_LIST = ["Off"] + ["%s ms" % x for x in range(100, 1100, 100)]
SAVE_LIST = ["Off", "Normal", "Super", "Deep"]
SCREV_LIST = ["Time (TO)", "Carrier (CO)", "Search (SE)"]
TAILCODE_LIST = ["55 Hz", "62.5 Hz"]
TAILPHASE_LIST = ["None", "120 Shift", "180 Shift", "240 Shift"]
TONERXEND_LIST = ["Off", "MDC-1200"]
TONE_LIST = ["1000 Hz", "1450 Hz", "1750 Hz", "2100 Hz"]
TOT_LIST = ["Off", "30 seconds", "60 seconds", "120 seconds", "240 seconds",
            "480 seconds"]
VOXD_LIST = ["%s seconds" % str(x / 10) for x in range(5, 21)]
WORKMODE_LIST = ["VFO Mode", "Channel Mode"]

ALL_SKEY_CHOICES = ["OFF",
                    "FM Radio",
                    "TX Power Level",
                    "Scan",
                    "Search",
                    "Flashlight",
                    "NOAA Weather",
                    "Monitor",
                    "PTT B",
                    "SOS",
                    "DTMF",
                    "REVERSE",
                    "REMOTE Scan"]

ALL_SKEY_VALUES = [0xFF,
                   0x07,
                   0x0A,
                   0x1C,
                   0x1D,
                   0x08,
                   0x0C,
                   0x05,
                   0x01,
                   0x03,
                   0x2A,
                   0x2D,
                   0x23]


def _enter_programming_mode(radio):
    serial = radio.pipe

    exito = False
    for i in range(0, 5):
        serial.write(radio._magic)
        ack = serial.read(1)

        try:
            if ack == CMD_ACK:
                exito = True
                break
        except Exception:
            LOG.debug("Attempt #%s, failed, trying again" % i)
            pass

    # check if we had EXITO
    if exito is False:
        msg = "The radio did not accept program mode after five tries.\n"
        msg += "Check you interface cable and power cycle your radio."
        raise errors.RadioError(msg)

    try:
        serial.write(b"F")
        ident = serial.read(8)
    except Exception:
        raise errors.RadioError("Error communicating with radio")

    if ident not in radio._fingerprint:
        LOG.debug(util.hexprint(ident))
        raise errors.RadioError("Radio returned unknown identification string")

    if radio.MODEL == "RT-470X":
        if ident in radio._fingerprint_pcb1:
            LOG.info("Radtel RT-470X - original pcb")
            radio.RT470X_ORIG = True
        elif ident in radio._fingerprint_pcb2:
            LOG.info("Radtel RT-470X - pcb2")
            radio.RT470X_ORIG = False


def _exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(b"E")
    except Exception:
        raise errors.RadioError("Radio refused to exit programming mode")


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"R" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]
    except Exception:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except Exception:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE_UP):
            status.cur = addr + radio.BLOCK_SIZE_UP
            radio.status_fn(status)
            _write_block(radio, addr, radio.BLOCK_SIZE_UP)

    _exit_programming_mode(radio)


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # if the two freqs are on the same Band this is not a split
            return False

    # if you get here is because the freq pairs are split
    return True


class JC8810base(chirp_common.CloneModeRadio):
    """MML JC-8810"""
    VENDOR = "MML"
    MODEL = "JC-8810base"
    BAUD_RATE = 57600
    NEEDS_COMPAT_SERIAL = False
    BLOCK_SIZE = 0x40
    BLOCK_SIZE_UP = 0x40

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=10.00),
                    chirp_common.PowerLevel("M", watts=8.00),
                    chirp_common.PowerLevel("L", watts=4.00)]

    VALID_BANDS = [(108000000, 136000000),
                   (136000000, 180000000),
                   (200000000, 260000000),
                   (330000000, 400000000),
                   (400000000, 520000000)]

    _magic = b"PROGRAMJC81U"
    _fingerprint = [b"\x00\x00\x00\x26\x00\x20\xD8\x04",
                    b"\x00\x00\x00\x42\x00\x20\xF0\x04",
                    b"\x00\x00\x00\x4A\x00\x20\xF8\x04"]

    _ranges = [
               (0x0000, 0x2000),
               (0x8000, 0x8040),
               (0x9000, 0x9040),
               (0xA000, 0xA140),
               (0xB000, 0xB300)
              ]
    _memsize = 0xB300
    _upper = 256
    _aninames = 30
    _mem_params = (_upper,  # number of channels
                   _aninames,  # number of aninames
                   )
    _valid_chars = chirp_common.CHARSET_ALPHANUMERIC + \
        "`~!@#$%^&*()-=_+[]\\{}|;':\",./<>?"

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = True
        rf.valid_name_length = 12
        rf.valid_characters = self._valid_chars
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_dtcs_codes = DTCS
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = [2.5, 5., 6.25, 8.33, 10., 12.5, 20., 25., 50.]
        rf.valid_bands = self.VALID_BANDS
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT % self._mem_params, self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = do_download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = data
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            do_upload(self)
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        _mem = self._memobj.memory[number - 1]

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Memory number
        mem.number = number

        if _mem.get_raw()[:1] == b"\xFF":
            mem.empty = True
            return mem

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # tx freq can be blank
        if _mem.txfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX freq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset != 0:
                if _split(self.get_features(), mem.freq, int(
                          _mem.txfreq) * 10):
                    mem.duplex = "split"
                    mem.offset = int(_mem.txfreq) * 10
                elif offset < 0:
                    mem.offset = abs(offset)
                    mem.duplex = "-"
                elif offset > 0:
                    mem.offset = offset
                    mem.duplex = "+"
            else:
                mem.offset = 0

        for char in _mem.name:
            if str(char) == "\xFF":
                char = " "  # may have 0xFF mid-name
            mem.name += str(char)
        mem.name = mem.name.rstrip()

        dtcs_pol = ["N", "N"]

        if _mem.txtone in [0, 0xFFFF]:
            txmode = ""
        elif _mem.txtone >= 0x0258:
            txmode = "Tone"
            mem.rtone = int(_mem.txtone) / 10.0
        elif _mem.txtone <= 0x0258:
            txmode = "DTCS"
            if _mem.txtone > 0x69:
                index = _mem.txtone - 0x6A
                dtcs_pol[0] = "R"
            else:
                index = _mem.txtone - 1
            mem.dtcs = DTCS[index]
        else:
            LOG.warn("Bug: txtone is %04x" % _mem.txtone)

        if _mem.rxtone in [0, 0xFFFF]:
            rxmode = ""
        elif _mem.rxtone >= 0x0258:
            rxmode = "Tone"
            mem.ctone = int(_mem.rxtone) / 10.0
        elif _mem.rxtone <= 0x0258:
            rxmode = "DTCS"
            if _mem.rxtone >= 0x6A:
                index = _mem.rxtone - 0x6A
                dtcs_pol[1] = "R"
            else:
                index = _mem.rxtone - 1
            mem.rx_dtcs = DTCS[index]
        else:
            LOG.warn("Bug: rxtone is %04x" % _mem.rxtone)

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        mem.dtcs_polarity = "".join(dtcs_pol)

        if not _mem.scan:
            mem.skip = "S"

        _levels = self.POWER_LEVELS
        if self.MODEL in ["A36plus", "A36plus_8w", "UV-A37", "AR-730"]:
            if _mem.txpower == TXPOWER_HIGH:
                mem.power = _levels[0]
            elif _mem.txpower == TXPOWER_LOW:
                mem.power = _levels[1]
            else:
                LOG.error('%s: get_mem: unhandled power level: 0x%02x' %
                          (mem.name, _mem.txpower))
        else:
            if _mem.txpower == TXPOWER_HIGH:
                mem.power = _levels[0]
            elif _mem.txpower == TXPOWER_MID:
                mem.power = _levels[1]
            elif _mem.txpower == TXPOWER_LOW:
                mem.power = _levels[2]
            else:
                LOG.error('%s: get_mem: unhandled power level: 0x%02x' %
                          (mem.name, _mem.txpower))

        mem.mode = _mem.narrow and "NFM" or "FM"

        mem.extra = RadioSettingGroup("Extra", "extra")

        # BCL (Busy Channel Lockout)
        rs = RadioSettingValueBoolean(_mem.bcl)
        rset = RadioSetting("bcl", "BCL", rs)
        mem.extra.append(rset)

        # PTT-ID
        rs = RadioSettingValueList(PTTID_LIST, PTTID_LIST[_mem.pttid])
        rset = RadioSetting("pttid", "PTT ID", rs)
        mem.extra.append(rset)

        # Signal (DTMF Encoder Group #)
        rs = RadioSettingValueList(PTTIDCODE_LIST, PTTIDCODE_LIST[_mem.scode])
        rset = RadioSetting("scode", "PTT ID Code", rs)
        mem.extra.append(rset)

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        if mem.empty:
            _mem.set_raw("\xff" * 32)
            return

        _mem.set_raw("\x00" * 16 + "\xFF" * 16)

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            _mem.txfreq.fill_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _namelength = self.get_features().valid_name_length
        for i in range(_namelength):
            try:
                _mem.name[i] = mem.name[i]
            except IndexError:
                _mem.name[i] = "\xFF"

        rxmode = txmode = ""
        if mem.tmode == "Tone":
            _mem.txtone = int(mem.rtone * 10)
            _mem.rxtone = 0
        elif mem.tmode == "TSQL":
            _mem.txtone = int(mem.ctone * 10)
            _mem.rxtone = int(mem.ctone * 10)
        elif mem.tmode == "DTCS":
            rxmode = txmode = "DTCS"
            _mem.txtone = DTCS.index(mem.dtcs) + 1
            _mem.rxtone = DTCS.index(mem.dtcs) + 1
        elif mem.tmode == "Cross":
            txmode, rxmode = mem.cross_mode.split("->", 1)
            if txmode == "Tone":
                _mem.txtone = int(mem.rtone * 10)
            elif txmode == "DTCS":
                _mem.txtone = DTCS.index(mem.dtcs) + 1
            else:
                _mem.txtone = 0
            if rxmode == "Tone":
                _mem.rxtone = int(mem.ctone * 10)
            elif rxmode == "DTCS":
                _mem.rxtone = DTCS.index(mem.rx_dtcs) + 1
            else:
                _mem.rxtone = 0
        else:
            _mem.rxtone = 0
            _mem.txtone = 0

        if txmode == "DTCS" and mem.dtcs_polarity[0] == "R":
            _mem.txtone += 0x69
        if rxmode == "DTCS" and mem.dtcs_polarity[1] == "R":
            _mem.rxtone += 0x69

        _mem.scan = mem.skip != "S"
        _mem.narrow = mem.mode == "NFM"

        _levels = self.POWER_LEVELS
        if self.MODEL in ["A36plus", "A36plus_8w", "UV-A37", "AR-730"]:
            if mem.power is None:
                _mem.txpower = TXPOWER_HIGH
            elif mem.power == _levels[0]:
                _mem.txpower = TXPOWER_HIGH
            elif mem.power == _levels[1]:
                _mem.txpower = TXPOWER_LOW
            else:
                LOG.error('%s: set_mem: unhandled power level: %s' %
                          (mem.name, mem.power))
        else:
            if mem.power is None:
                _mem.txpower = TXPOWER_HIGH
            elif mem.power == _levels[0]:
                _mem.txpower = TXPOWER_HIGH
            elif mem.power == _levels[1]:
                _mem.txpower = TXPOWER_MID
            elif mem.power == _levels[2]:
                _mem.txpower = TXPOWER_LOW
            else:
                LOG.error('%s: set_mem: unhandled power level: %s' %
                          (mem.name, mem.power))

        for setting in mem.extra:
            if setting.get_name() == "scramble_type":
                setattr(_mem, setting.get_name(), int(setting.value) + 8)
                setattr(_mem, "scramble_type2", int(setting.value) + 8)
            else:
                setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _dtmf = self._memobj.dtmf
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        group = RadioSettings(basic)

        # Menu 12: TOT
        rs = RadioSettingValueList(TOT_LIST, TOT_LIST[_settings.tot])
        rset = RadioSetting("tot", "Time Out Timer", rs)
        basic.append(rset)

        # Menu 00: SQL
        rs = RadioSettingValueInteger(0, 9, _settings.sql)
        rset = RadioSetting("sql", "Squelch Level", rs)
        basic.append(rset)

        # Menu 13: VOX
        rs = RadioSettingValueList(OFF1TO9_LIST, OFF1TO9_LIST[_settings.vox])
        rset = RadioSetting("vox", "VOX", rs)
        basic.append(rset)

        # Menu 39: VOX DELAY
        rs = RadioSettingValueList(VOXD_LIST, VOXD_LIST[_settings.voxd])
        rset = RadioSetting("voxd", "VOX Delay", rs)
        basic.append(rset)

        # Menu 15: VOICE
        rs = RadioSettingValueBoolean(_settings.voice)
        rset = RadioSetting("voice", "Voice Prompts", rs)
        basic.append(rset)

        if self.MODEL not in ["A36plus", "A36plus_8w", "UV-A37", "AR-730"]:
            # Menu 17: LANGUAGE
            rs = RadioSettingValueList(LANGUAGE_LIST,
                                       LANGUAGE_LIST[_settings.language])
            rset = RadioSetting("language", "Voice", rs)
            basic.append(rset)

        # Menu 23: ABR
        rs = RadioSettingValueList(ABR_LIST, ABR_LIST[_settings.abr])
        rset = RadioSetting("abr", "Auto BackLight", rs)
        basic.append(rset)

        # Work Mode A
        rs = RadioSettingValueList(WORKMODE_LIST,
                                   WORKMODE_LIST[_settings.vfomra])
        rset = RadioSetting("vfomra", "Work Mode A", rs)
        basic.append(rset)

        # Work Mode B
        rs = RadioSettingValueList(WORKMODE_LIST,
                                   WORKMODE_LIST[_settings.vfomrb])
        rset = RadioSetting("vfomrb", "Work Mode B", rs)
        basic.append(rset)

        # Menu 19: SC-REV
        rs = RadioSettingValueList(SCREV_LIST, SCREV_LIST[_settings.screv])
        rset = RadioSetting("screv", "Scan Resume Method", rs)
        basic.append(rset)

        # Menu 10: SAVE
        rs = RadioSettingValueList(SAVE_LIST,
                                   SAVE_LIST[_settings.save])
        rset = RadioSetting("save", "Battery Save Mode", rs)
        basic.append(rset)

        # Menu 42: MDF-A
        rs = RadioSettingValueList(MDF_LIST, MDF_LIST[_settings.mdfa])
        rset = RadioSetting("mdfa", "Memory Display Format A", rs)
        basic.append(rset)

        # Menu 43: MDF-B
        rs = RadioSettingValueList(MDF_LIST, MDF_LIST[_settings.mdfb])
        rset = RadioSetting("mdfb", "Memory Display Format B", rs)
        basic.append(rset)

        # Menu 33: DTMFST (DTMF ST)
        rs = RadioSettingValueList(DTMFST_LIST, DTMFST_LIST[_settings.dtmfst])
        rset = RadioSetting("dtmfst", "DTMF Side Tone", rs)
        basic.append(rset)

        # Menu 37: PTT-LT
        rs = RadioSettingValueList(PTTLT_LIST, PTTLT_LIST[_settings.pttlt])
        rset = RadioSetting("pttlt", "PTT Delay", rs)
        basic.append(rset)

        rs = RadioSettingValueInteger(1, 60, _settings.ani + 1)
        rset = RadioSetting("ani", "ANI", rs)
        basic.append(rset)

        # Menu 20: PF2
        def apply_skey2s_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) + " from list")
            val = str(setting.value)
            index = SKEY2S_CHOICES.index(val)
            val = SKEY2S_VALUES[index]
            obj.set_value(val)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
            unwanted = [9, 10, 11, 12]
        elif self.MODEL in ["UV-A37", "AR-730"]:
            unwanted = [0, 5, 7, 9, 10, 11, 12]
        elif self.MODEL in ["A36plus", "A36plus_8w"]:
            unwanted = [0, 5, 7, 9, 10, 11]
        else:
            unwanted = []
        SKEY2S_CHOICES = ALL_SKEY_CHOICES.copy()
        SKEY2S_VALUES = ALL_SKEY_VALUES.copy()
        for ele in sorted(unwanted, reverse=True):
            del SKEY2S_CHOICES[ele]
            del SKEY2S_VALUES[ele]

        if _settings.skey2_sp in SKEY2S_VALUES:
            idx = SKEY2S_VALUES.index(_settings.skey2_sp)
        else:
            idx = SKEY2S_VALUES.index(0x07)  # default FM
        rs = RadioSettingValueList(SKEY2S_CHOICES, SKEY2S_CHOICES[idx])
        rset = RadioSetting("skey2_sp", "PF2 Key (Short Press)", rs)
        rset.set_apply_callback(apply_skey2s_listvalue, _settings.skey2_sp)
        basic.append(rset)

        # Menu 21: PF2 LONG PRESS
        def apply_skey2l_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) + " from list")
            val = str(setting.value)
            index = SKEY2L_CHOICES.index(val)
            val = SKEY2L_VALUES[index]
            obj.set_value(val)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
            unwanted = [8, 9, 10, 11, 12]
        elif self.MODEL in ["UV-A37", "AR-730"]:
            unwanted = [0, 5, 7, 8, 10, 11, 12]
        elif self.MODEL in ["A36plus", "A36plus_8w"]:
            unwanted = [0, 5, 7, 8, 11, 12]
        else:
            unwanted = []
        SKEY2L_CHOICES = ALL_SKEY_CHOICES.copy()
        SKEY2L_VALUES = ALL_SKEY_VALUES.copy()
        for ele in sorted(unwanted, reverse=True):
            del SKEY2L_CHOICES[ele]
            del SKEY2L_VALUES[ele]

        if _settings.skey2_lp in SKEY2L_VALUES:
            idx = SKEY2L_VALUES.index(_settings.skey2_lp)
        else:
            idx = SKEY2L_VALUES.index(0x1D)  # default Search
        rs = RadioSettingValueList(SKEY2L_CHOICES, SKEY2L_CHOICES[idx])
        rset = RadioSetting("skey2_lp", "PF2 Key (Long Press)", rs)
        rset.set_apply_callback(apply_skey2l_listvalue, _settings.skey2_lp)
        basic.append(rset)

        # Menu 22: PF3
        def apply_skey3s_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) + " from list")
            val = str(setting.value)
            index = SKEY3S_CHOICES.index(val)
            val = SKEY3S_VALUES[index]
            obj.set_value(val)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
            unwanted = [8, 9, 10, 11, 12]
        elif self.MODEL in ["UV-A37", "AR-730"]:
            unwanted = [0, 5, 7, 8, 9, 10, 11, 12]
        elif self.MODEL in ["A36plus", "A36plus_8w"]:
            unwanted = [0, 5, 7, 8, 11]
        else:
            unwanted = []
        SKEY3S_CHOICES = ALL_SKEY_CHOICES.copy()
        SKEY3S_VALUES = ALL_SKEY_VALUES.copy()
        for ele in sorted(unwanted, reverse=True):
            del SKEY3S_CHOICES[ele]
            del SKEY3S_VALUES[ele]

        if _settings.skey3_sp in SKEY3S_VALUES:
            idx = SKEY3S_VALUES.index(_settings.skey3_sp)
        else:
            idx = SKEY3S_VALUES.index(0x0C)  # default NOAA
        rs = RadioSettingValueList(SKEY3S_CHOICES, SKEY3S_CHOICES[idx])
        rset = RadioSetting("skey3_sp", "PF3 Key (Short Press)", rs)
        rset.set_apply_callback(apply_skey3s_listvalue, _settings.skey3_sp)
        basic.append(rset)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
            # Menu 24: PF3 LONG PRESS (RT-470L)
            def apply_skey3l_listvalue(setting, obj):
                LOG.debug("Setting value: " + str(setting.value) +
                          " from list")
                val = str(setting.value)
                index = SKEY3L_CHOICES.index(val)
                val = SKEY2L_VALUES[index]
                obj.set_value(val)

            if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
                unwanted = [8, 9, 10, 11, 12]
            else:
                unwanted = []
            SKEY3L_CHOICES = ALL_SKEY_CHOICES.copy()
            SKEY3L_VALUES = ALL_SKEY_VALUES.copy()
            for ele in sorted(unwanted, reverse=True):
                del SKEY3L_CHOICES[ele]
                del SKEY3L_VALUES[ele]

            if _settings.skey3_lp in SKEY3L_VALUES:
                idx = SKEY3L_VALUES.index(_settings.skey3_lp)
            else:
                idx = SKEY3L_VALUES.index(0x1D)  # default SEARCH
            rs = RadioSettingValueList(SKEY3L_CHOICES, SKEY3L_CHOICES[idx])
            rset = RadioSetting("skey3_lp", "PF3 Key (Long Press)", rs)
            rset.set_apply_callback(apply_skey3l_listvalue,
                                    _settings.skey3_lp)
            basic.append(rset)

        # Menu 25: TOP KEY (RT-470L)
        def apply_skeytop_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) +
                      " from list")
            val = str(setting.value)
            index = SKEYTOP_CHOICES.index(val)
            val = SKEYTOP_VALUES[index]
            obj.set_value(val)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X", "RT-470"]:
            unwanted = [8, 9, 10, 11, 12]
        elif self.MODEL in ["UV-A37", "AR-730"]:
            unwanted = [0, 5, 7, 8, 9, 10, 11, 12]
        elif self.MODEL in ["A36plus", "A36plus_8w"]:
            unwanted = [0, 5, 7, 8, 11]
        else:
            unwanted = []
        SKEYTOP_CHOICES = ALL_SKEY_CHOICES.copy()
        SKEYTOP_VALUES = ALL_SKEY_VALUES.copy()
        for ele in sorted(unwanted, reverse=True):
            del SKEYTOP_CHOICES[ele]
            del SKEYTOP_VALUES[ele]

        if _settings.topkey_sp in SKEYTOP_VALUES:
            idx = SKEYTOP_VALUES.index(_settings.topkey_sp)
        else:
            idx = SKEYTOP_VALUES.index(0x1D)  # default SEARCH
        rs = RadioSettingValueList(SKEYTOP_CHOICES, SKEYTOP_CHOICES[idx])
        rset = RadioSetting("topkey_sp", "Top Key (Short Press)", rs)
        rset.set_apply_callback(apply_skeytop_listvalue,
                                _settings.topkey_sp)
        basic.append(rset)

        # Mneu 36: TONE
        rs = RadioSettingValueList(TONE_LIST, TONE_LIST[_settings.tone])
        rset = RadioSetting("tone", "Tone-burst Frequency", rs)
        basic.append(rset)

        # Mneu 29: POWER ON MSG
        rs = RadioSettingValueList(PONMSG_LIST, PONMSG_LIST[_settings.ponmsg])
        rset = RadioSetting("ponmsg", "Power On Message", rs)
        basic.append(rset)

        if self.MODEL in ["HI-8811", "RT-470L", "RT-470X"]:
            rs = RadioSettingValueList(TAILCODE_LIST,
                                       TAILCODE_LIST[_settings.tailcode])
            rset = RadioSetting("tailcode", "Tail Code", rs)
            basic.append(rset)

        # Menu 46: STE
        rs = RadioSettingValueBoolean(_settings.ste)
        rset = RadioSetting("ste", "Squelch Tail Eliminate (HT to HT)", rs)
        basic.append(rset)

        # Menu 40: RP-STE
        rs = RadioSettingValueList(RPSTE_LIST, RPSTE_LIST[_settings.rpste])
        rset = RadioSetting("rpste", "Squelch Tail Eliminate (repeater)", rs)
        basic.append(rset)

        # Menu 41: RPT-RL
        rs = RadioSettingValueList(RPSTE_LIST, RPSTE_LIST[_settings.rptrl])
        rset = RadioSetting("rptrl", "STE Repeater Delay", rs)
        basic.append(rset)

        # Menu 38: MENU EXIT TIME
        rs = RadioSettingValueList(MENUQUIT_LIST,
                                   MENUQUIT_LIST[_settings.menuquit])
        rset = RadioSetting("menuquit", "Menu Auto Quit", rs)
        basic.append(rset)

        # Menu 34: AUTOLOCK
        rs = RadioSettingValueList(AUTOLK_LIST, AUTOLK_LIST[_settings.autolk])
        rset = RadioSetting("autolk", "Key Auto Lock", rs)
        basic.append(rset)

        # Menu 28: CDCSS SAVE MODE
        rs = RadioSettingValueList(QTSAVE_LIST, QTSAVE_LIST[_settings.qtsave])
        rset = RadioSetting("qtsave", "QT Save Type", rs)
        basic.append(rset)

        # Menu 45: TX-A/B
        rs = RadioSettingValueList(DUALTX_LIST, DUALTX_LIST[_settings.dualtx])
        rset = RadioSetting("dualtx", "Dual TX", rs)
        basic.append(rset)

        # Menu 47: AL-MODE
        rs = RadioSettingValueList(ALMODE_LIST, ALMODE_LIST[_settings.almode])
        rset = RadioSetting("almode", "Alarm Mode", rs)
        basic.append(rset)

        # Menu 11: ROGER
        # ==========
        # Notice to developers:
        # The RT-470 v1.22 firmware expanded the ROGER menu with an additional
        # choice, 'TONE1200'. RT-470 radios with a firmware version prior to
        #  v1.22 will not honor the ROGER menu's 'TONE1200' choice in CHIRP.
        # ==========
        rs = RadioSettingValueList(ROGER_LIST, ROGER_LIST[_settings.roger])
        rset = RadioSetting("roger", "Roger", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.alarmsound)
        rset = RadioSetting("alarmsound", "Alarm Sound", rs)
        basic.append(rset)

        # Menu 44: TDR
        rs = RadioSettingValueBoolean(_settings.tdr)
        rset = RadioSetting("tdr", "TDR", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(not _settings.fmradio)
        rset = RadioSetting("fmradio", "FM Radio", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.kblock)
        rset = RadioSetting("kblock", "KB Lock", rs)
        basic.append(rset)

        # Menu 16: BEEP PROMPT
        rs = RadioSettingValueBoolean(_settings.beep)
        rset = RadioSetting("beep", "Beep", rs)
        basic.append(rset)

        if self.MODEL not in ["A36plus", "A36plus_8w", "UV-A37", "AR-730"]:
            # Menu 48: RX END TAIL
            rs = RadioSettingValueList(TONERXEND_LIST,
                                       TONERXEND_LIST[_settings.rxendtail])
            rset = RadioSetting("rxendtail", "Tone RX End", rs)
            basic.append(rset)

        dtmf = RadioSettingGroup("dtmf", "DTMF Settings")
        group.append(dtmf)

        def apply_code(setting, obj, length):
            code = []
            for j in range(0, length):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.code = code

        for i in range(0, 15):
            _codeobj = self._memobj.pttid[i].code
            _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            rs = RadioSettingValueString(0, 5, _code, False)
            rs.set_charset(DTMF_CHARS)
            rset = RadioSetting("pttid/%i.code" % i,
                                "PTT-ID Code %i" % (i + 1), rs)
            rset.set_apply_callback(apply_code, self._memobj.pttid[i], 5)
            dtmf.append(rset)

        rs = RadioSettingValueList(DTMFSPEED_LIST,
                                   DTMFSPEED_LIST[_dtmf.dtmfon])
        rset = RadioSetting("dtmf.dtmfon", "DTMF Speed (on)", rs)
        dtmf.append(rset)

        rs = RadioSettingValueList(DTMFSPEED_LIST,
                                   DTMFSPEED_LIST[_dtmf.dtmfoff])
        rset = RadioSetting("dtmf.dtmfoff", "DTMF Speed (off)", rs)
        dtmf.append(rset)

        rs = RadioSettingValueList(PTTID_LIST, PTTID_LIST[_settings.pttid])
        rset = RadioSetting("pttid", "PTT ID", rs)
        dtmf.append(rset)

        ani = RadioSettingGroup("ani", "ANI Code List Settings")
        group.append(ani)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in self._valid_chars:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        end = 60 - self._aninames  # end of immutable ANI names
        for i in range(0, end):
            _codeobj = self._memobj.anicodes[i].code
            _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            rs = RadioSettingValueString(0, 3, _code, False)
            rs.set_charset(DTMF_CHARS)
            rset = RadioSetting("anicodes/%i.code" % i,
                                "ANI Code %i (NUM.%i)" % (i + 1, i + 1), rs)
            rset.set_apply_callback(apply_code, self._memobj.anicodes[i], 3)
            ani.append(rset)

        for i in range(end, 60):
            _codeobj = self._memobj.anicodes[i].code
            _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            rs = RadioSettingValueString(0, 3, _code, False)
            rs.set_charset(DTMF_CHARS)
            rset = RadioSetting("anicodes/%i.code" % (i),
                                "ANI Code %i" % (i + 1), rs)
            rset.set_apply_callback(apply_code,
                                    self._memobj.anicodes[i], 3)
            ani.append(rset)

            _nameobj = self._memobj.aninames[i - end].name
            rs = RadioSettingValueString(0, 10, _filter(_nameobj))
            rset = RadioSetting("aninames/%i.name" % (i - end),
                                "ANI Code %i Name" % (i + 1), rs)
            ani.append(rset)

        if self.MODEL in ["A36plus", "A36plus_8w"]:
            custom = RadioSettingGroup("custom", "Custom Channel Names")
            group.append(custom)

            for i in range(0, 30):
                _nameobj = self._memobj.customnames[i].name
                rs = RadioSettingValueString(0, 10, _filter(_nameobj))
                rset = RadioSetting("customnames/%i.name" % i,
                                    "Custom Name %i" % (i + 1), rs)
                custom.append(rset)

        if self.MODEL in ["A36plus", "A36plus_8w"]:
            # Menu 21: RX END TAIL
            rs = RadioSettingValueList(TONERXEND_LIST,
                                       TONERXEND_LIST[_settings.skey3_lp])
            rset = RadioSetting("skey3_lp", "RX End Tail", rs)
            basic.append(rset)

            # Menu 23: TAIL PHASE
            rs = RadioSettingValueList(TAILPHASE_LIST,
                                       TAILPHASE_LIST[_settings.rxendtail])
            rset = RadioSetting("rxendtail", "Tail Phase", rs)
            basic.append(rset)

        return group

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "ani":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "fmradio":
                        setattr(obj, setting, not int(element.value))
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception as e:
                    LOG.debug(element.get_name(), e)
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RT470Radio(JC8810base):
    """Radtel RT-470"""
    VENDOR = "Radtel"
    MODEL = "RT-470"

    # ==========
    # Notice to developers:
    # The RT-470 support in this driver is currently based upon...
    # - v1.25a firmware (original pcb)
    # - v2.11a firmware (pcb2)
    # ==========

    # original pcb
    _fingerprint_pcb1 = [b"\x00\x00\x00\x26\x00\x20\xD8\x04",
                         b"\x00\x00\x00\x42\x00\x20\xF0\x04",
                         b"\x00\x00\x00\x4A\x00\x20\xF8\x04",
                         b"\x00\x00\x00\x3A\x00\x20\xE8\x04",  # fw 1.25A
                         ]

    # pcb 2
    _fingerprint_pcb2 = [b"\x00\x00\x00\x2C\x00\x20\xD8\x04",  # fw v2.11A
                         ]

    _fingerprint = _fingerprint_pcb1 + _fingerprint_pcb2

    VALID_BANDS = [(16000000, 100000000),
                   (100000000, 136000000),
                   (136000000, 200000000),
                   (200000000, 300000000),
                   (300000000, 400000000),
                   (400000000, 560000000),
                   (740000000, 1000000000),
                   ]


@directory.register
class RT470LRadio(JC8810base):
    """Radtel RT-470L"""
    VENDOR = "Radtel"
    MODEL = "RT-470L"

    # ==========
    # Notice to developers:
    # The RT-470 support in this driver is currently based upon v1.17 firmware.
    # ==========

    _fingerprint = [b"\x00\x00\x00\xfe\x00\x20\xAC\x04",
                    b"\x00\x00\x00\x20\x00\x20\xCC\x04",
                    b"\x00\x00\x00\x20\x00\x20\x07\x00"]

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=5.00),
                    chirp_common.PowerLevel("M", watts=4.00),
                    chirp_common.PowerLevel("L", watts=2.00)]

    VALID_BANDS = [(108000000, 136000000),
                   (136000000, 179000000),
                   (220000000, 260000000),
                   (330000000, 400000000),
                   (400000000, 520000000)]


@directory.register
class RT470XRadio(RT470LRadio):
    """Radtel RT-470X"""
    VENDOR = "Radtel"
    MODEL = "RT-470X"

    # ==========
    # Notice to developers:
    # The RT-470X support in this driver is currently based upon...
    # - v1.18a firmware (original pcb)
    # - v2.13a firmware (pcb2)
    # ==========

    # original pcb
    _fingerprint_pcb1 = [b"\x00\x00\x00\x20\x00\x20\xCC\x04",
                         ]

    # pcb 2
    _fingerprint_pcb2 = [b"\x00\x00\x00\x2C\x00\x20\xD8\x04",  # fw v2.10A
                         b"\x00\x00\x00\x36\x00\x20\xDC\x04",  # fw v2.13A
                         ]

    _fingerprint = _fingerprint_pcb1 + _fingerprint_pcb2

    RT470X_ORIG = False

    VALID_BANDS = [(100000000, 136000000),
                   (136000000, 200000000),
                   (200000000, 300000000),
                   (300000000, 400000000),
                   (400000000, 560000000)]


@directory.register
class HI8811Radio(RT470LRadio):
    """Hiroyasu HI-8811"""
    VENDOR = "Hiroyasu"
    MODEL = "HI-8811"

    # ==========
    # Notice to developers:
    # The HI-8811 support in this driver is currently based upon...
    # - v1.17 firmware (original pcb)
    # - v2.00 firmware (pcb2)
    # ==========

    # original pcb
    _fingerprint_pcb1 = [b"\x00\x00\x00\xfe\x00\x20\xAC\x04",
                         b"\x00\x00\x00\x20\x00\x20\xCC\x04",  # fw v1.17
                         b"\x00\x00\x00\x20\x00\x20\x07\x00",
                         ]

    # pcb 2
    _fingerprint_pcb2 = [b"\x00\x00\x00\x28\x00\x20\xD4\x04",  # fw v2.00
                         b"\x00\x00\x00\x28\x00\x20\x07\x00",  # fw v2.00
                         ]

    _fingerprint = _fingerprint_pcb1 + _fingerprint_pcb2


@directory.register
class UVA37Radio(JC8810base):
    """Anysecu UV-A37"""
    VENDOR = "Anysecu"
    MODEL = "UV-A37"

    # ==========
    # Notice to developers:
    # The UV-A37 support in this driver is currently based upon v1.24
    # firmware.
    # ==========

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=5.00),
                    chirp_common.PowerLevel("L", watts=1.00)]

    VALID_BANDS = [(108000000, 136000000),
                   (136000000, 174000000),
                   (200000000, 260000000),
                   (350000000, 390000000),
                   (400000000, 520000000)]

    _magic = b"PROGRAMJC37U"
    _fingerprint = [b"\x00\x00\x00\xE4\x00\x20\x94\x04",
                    b"\x00\x00\x00\xE8\x00\x20\x98\x04"]

    _ranges = [
               (0x0000, 0x2000),
               (0x8000, 0x8040),
               (0x9000, 0x9040),
               (0xA000, 0xA140),
               (0xB000, 0xB440)
              ]
    _memsize = 0xB440


@directory.register
class A36plusRadio(JC8810base):
    """Talkpod A36plus"""
    VENDOR = "Talkpod"
    MODEL = "A36plus"

    # ==========
    # Notice to developers:
    # The A36plus support in this driver is currently based upon v1.22
    # firmware.
    # ==========

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=5.00),
                    chirp_common.PowerLevel("L", watts=1.00)]

    VALID_BANDS = [(108000000, 136000000),
                   (136000000, 180000000),
                   (200000000, 260000000),
                   (350000000, 400000000),
                   (400000000, 520000000),
                   ]

    _magic = b"PROGRAMJC37U"
    _fingerprint = [b"\x00\x00\x00\x42\x00\x20\xF0\x04",
                    b"\x00\x00\x00\x5A\x00\x20\x08\x05",  # fw 1.18
                    b"\x00\x00\x00\x9E\x00\x20\x0C\x05",  # fw 1.22
                    b"\x00\x00\x00\xFA\x00\x20\x40\x05",  # fw 1.4
                    ]

    _ranges = [
               (0x0000, 0x4000),
               (0x8000, 0x8040),
               (0x9000, 0x9040),
               (0xA000, 0xA140),
               (0xB000, 0xB440)
              ]
    _memsize = 0xB440
    _upper = 512  # fw 1.22 expands from 256 to 512 channels
    _aninames = 10
    _mem_params = (_upper,  # number of channels
                   _aninames,  # number of aninames
                   )

    def process_mmap(self):
        mem_format = MEM_FORMAT % self._mem_params + MEM_FORMAT_A36PLUS
        self._memobj = bitwise.parse(mem_format, self._mmap)


@directory.register
class A36plus8wRadio(A36plusRadio):
    """Talkpod A36plus8w"""
    VENDOR = "Talkpod"
    MODEL = "A36plus_8w"

    # ==========
    # Notice to developers:
    # The A36plus 8w support in this driver is currently based upon v1.4
    # firmware.
    # ==========

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=8.00),
                    chirp_common.PowerLevel("L", watts=1.00)]

    _fingerprint = [b"\x00\x00\x00\xFA\x00\x20\x40\x05",  # fw 1.4
                    ]


@directory.register
class AR730Radio(UVA37Radio):
    """Abbree AR730"""
    VENDOR = "Abbree"
    MODEL = "AR-730"

    # ==========
    # Notice to developers:
    # The AR-730 support in this driver is currently based upon v1.24
    # firmware.
    # ==========

    POWER_LEVELS = [chirp_common.PowerLevel("H", watts=5.00),
                    chirp_common.PowerLevel("L", watts=1.00)]

    VALID_BANDS = [(108000000, 136000000),
                   (136000000, 180000000),
                   (200000000, 260000000),
                   (350000000, 390000000),
                   (400000000, 520000000)]



# Copyright 2016:
# * Jim Unroe KC9HI, <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import baofeng_common as bfc
from chirp import chirp_common, directory
from chirp import bitwise
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettingValueFloat, RadioSettings, \
    InvalidValueError

LOG = logging.getLogger(__name__)

# #### MAGICS #########################################################

# Baofeng WP970I magic string
MSTRING_WP970I = b"\x50\xBB\xFF\x20\x14\x04\x13"

# Baofeng UV-9G magic string
MSTRING_UV9G = b"\x50\xBB\xFF\x20\x12\x05\x25"

# Baofeng UV-S9X3 magic string
MSTRING_UVS9X3 = b"\x50\xBB\xFF\x20\x12\x07\x25"


DTMF_CHARS = "0123456789 *#ABCD"
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0]

LIST_AB = ["A", "B"]
LIST_ALMOD = ["Site", "Tone", "Code"]
LIST_BANDWIDTH = ["Wide", "Narrow"]
LIST_COLOR = ["Off", "Blue", "Orange", "Purple"]
LIST_DTMFSPEED = ["%s ms" % x for x in range(50, 2010, 10)]
LIST_DTMFST = ["Off", "DT-ST", "ANI-ST", "DT+ANI"]
LIST_MODE = ["Channel", "Name", "Frequency"]
LIST_OFF1TO9 = ["Off"] + list("123456789")
LIST_OFF1TO10 = LIST_OFF1TO9 + ["10"]
LIST_OFFAB = ["Off"] + LIST_AB
LIST_RESUME = ["TO", "CO", "SE"]
LIST_PONMSG = ["Full", "Message"]
LIST_PTTID = ["Off", "BOT", "EOT", "Both"]
LIST_SCODE = ["%s" % x for x in range(1, 16)]
LIST_RPSTE = ["Off"] + ["%s" % x for x in range(1, 11)]
LIST_SAVE = ["Off", "1:1", "1:2", "1:3", "1:4"]
LIST_SHIFTD = ["Off", "+", "-"]
LIST_STEDELAY = ["Off"] + ["%s ms" % x for x in range(100, 1100, 100)]
LIST_STEP = [str(x) for x in STEPS]
LIST_TIMEOUT = ["%s sec" % x for x in range(15, 615, 15)]
LIST_TXPOWER = ["High", "Mid", "Low"]
LIST_VOICE = ["Off", "English", "Chinese"]
LIST_WORKMODE = ["Frequency", "Channel"]

TXP_CHOICES = ["High", "Low"]
TXP_VALUES = [0x00, 0x02]


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""

    if len(data) > 0x2008:
        rid = data[0x2008:0x2010]
        return rid.startswith(cls.MODEL.encode())
    elif len(data) == 0x2008:
        rid = data[0x1EF0:0x1EF7]
        return rid in cls._fileid
    else:
        return False


class WP970I(bfc.BaofengCommonHT):
    """Baofeng WP970I"""
    VENDOR = "Baofeng"
    MODEL = "WP970I"
    NEEDS_COMPAT_SERIAL = False

    _tri_band = False
    _fileid = []
    _magic = [MSTRING_WP970I, ]
    _magic_response_length = 8
    _fw_ver_start = 0x1EF0
    _recv_block_size = 0x40
    _mem_size = 0x2000
    _ack_block = True

    _ranges = [(0x0000, 0x0DF0),
               (0x0E00, 0x1800),
               (0x1EE0, 0x1EF0),
               (0x1F60, 0x1F70),
               (0x1F80, 0x1F90),
               (0x1FC0, 0x1FD0)]
    _send_block_size = 0x10

    MODES = ["NFM", "FM"]
    VALID_CHARS = chirp_common.CHARSET_ALPHANUMERIC + \
        "!@#$%^&*()+-=[]:\";'<>?,./"
    LENGTH_NAME = 6
    SKIP_VALUES = ["", "S"]
    DTCS_CODES = tuple(sorted(chirp_common.DTCS_CODES + (645,)))
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Med",  watts=3.00),
                    chirp_common.PowerLevel("Low",  watts=1.00)]
    _vhf_range = (130000000, 180000000)
    _vhf2_range = (200000000, 260000000)
    _uhf_range = (400000000, 521000000)
    VALID_BANDS = [_vhf_range,
                   _uhf_range]
    PTTID_LIST = LIST_PTTID
    SCODE_LIST = LIST_SCODE

    MEM_FORMAT = """
    // #seekto 0x0000;
    struct {
      lbcd rxfreq[4];
      lbcd txfreq[4];
      ul16 rxtone;
      ul16 txtone;
      u8 unused1:3,
         isuhf:1,
         scode:4;
      u8 unknown1:7,
         txtoneicon:1;
      u8 mailicon:3,
         unknown2:3,
         lowpower:2;
      u8 unknown3:1,
         wide:1,
         unknown4:2,
         bcl:1,
         scan:1,
         pttid:2;
    } memory[128];

    #seekto 0x0B00;
    struct {
      u8 code[5];
      u8 unused[11];
    } pttid[15];

    #seekto 0x0CAA;
    struct {
      u8 code[5];
      u8 unused1:6,
         aniid:2;
      u8 unknown[2];
      u8 dtmfon;
      u8 dtmfoff;
    } ani;

    #seekto 0x0E20;
    struct {
      u8 squelch;
      u8 step;
      u8 unknown_e22;
      u8 save;
      u8 vox;
      u8 unknown_e25;
      u8 abr;
      u8 tdr;
      u8 beep;
      u8 timeout;
      u8 unknown_e2a[4];
      u8 voice;
      u8 unknown_e2f;
      u8 dtmfst;
      u8 unknown_e31;
      u8 unknown_e32:6,
         screv:2;
      u8 pttid;
      u8 pttlt;
      u8 mdfa;
      u8 mdfb;
      u8 bcl;
      u8 autolk;
      u8 sftd;
      u8 unknown_e3a[3];
      u8 wtled;
      u8 rxled;
      u8 txled;
      u8 almod;
      u8 band;
      u8 tdrab;
      u8 ste;
      u8 rpste;
      u8 rptrl;
      u8 ponmsg;
      u8 roger;
      u8 rogerrx;
      u8 tdrch;
      u8 displayab:1,
         unknown_e4a1:2,
         fmradio:1,
         alarm:1,
         unknown_e4a2:1,
         reset:1,
         menu:1;
      u8 unknown_e4b:6,
         singleptt:1,
         vfomrlock:1;
      u8 workmode;
      u8 keylock;
    } settings;

    #seekto 0x0E76;
    struct {
      u8 unused1:1,
         mrcha:7;
      u8 unused2:1,
         mrchb:7;
    } wmchannel;

    struct vfo {
      u8 unknown0[8];
      u8 freq[8];
      u8 offset[6];
      ul16 rxtone;
      ul16 txtone;
      u8 unused1:7,
         band:1;
      u8 unknown3;
      u8 unused2:2,
         sftd:2,
         scode:4;
      u8 unknown4;
      u8 unused3:1,
         step:3,
         unused4:4;
      u8 unused5:1,
         widenarr:1,
         unused6:4,
         txpower3:2;
    };

    #seekto 0x0F00;
    struct {
      struct vfo a;
      struct vfo b;
    } vfo;

    #seekto 0x0F4E;
    u16 fm_presets;

    #seekto 0x1000;
    struct {
      char name[7];
      u8 unknown1[9];
    } names[128];

    #seekto 0x1ED0;
    struct {
      char line1[7];
      char line2[7];
    } sixpoweron_msg;

    #seekto 0x1EE0;
    struct {
      char line1[7];
      char line2[7];
    } poweron_msg;

    #seekto 0x1EF0;
    struct {
      char line1[7];
      char line2[7];
    } firmware_msg;

    struct squelch {
      u8 sql0;
      u8 sql1;
      u8 sql2;
      u8 sql3;
      u8 sql4;
      u8 sql5;
      u8 sql6;
      u8 sql7;
      u8 sql8;
      u8 sql9;
    };

    #seekto 0x1F60;
    struct {
      struct squelch vhf;
      u8 unknown1[6];
      u8 unknown2[16];
      struct squelch uhf;
    } squelch;

    struct limit {
      u8 enable;
      bbcd lower[2];
      bbcd upper[2];
    };

    #seekto 0x1FC0;
    struct {
      struct limit vhf;
      struct limit uhf;
      struct limit vhf2;
    } limits;

    """

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is a beta version.\n'
             '\n'
             'Please save an unedited copy of your first successful\n'
             'download to a CHIRP Radio Images(*.img) file.'
             )
        rp.pre_download = _(
            "Follow these instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        rf = bfc.BaofengCommonHT.get_features(self)
        rf.valid_tuning_steps = STEPS
        return rf

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(self.MEM_FORMAT, self._mmap)

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic Settings")
        advanced = RadioSettingGroup("advanced", "Advanced Settings")
        other = RadioSettingGroup("other", "Other Settings")
        work = RadioSettingGroup("work", "Work Mode Settings")
        fm_preset = RadioSettingGroup("fm_preset", "FM Preset")
        dtmfe = RadioSettingGroup("dtmfe", "DTMF Encode Settings")
        service = RadioSettingGroup("service", "Service Settings")
        top = RadioSettings(basic, advanced, other, work, fm_preset, dtmfe,
                            service)

        # Basic settings
        if _mem.settings.squelch > 0x09:
            val = 0x00
        else:
            val = _mem.settings.squelch
        rs = RadioSetting("settings.squelch", "Squelch",
                          RadioSettingValueList(
                              LIST_OFF1TO9, LIST_OFF1TO9[val]))
        basic.append(rs)

        if _mem.settings.save > 0x04:
            val = 0x00
        else:
            val = _mem.settings.save
        rs = RadioSetting("settings.save", "Battery Saver",
                          RadioSettingValueList(
                              LIST_SAVE, LIST_SAVE[val]))
        basic.append(rs)

        if _mem.settings.vox > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.vox
        rs = RadioSetting("settings.vox", "Vox",
                          RadioSettingValueList(
                              LIST_OFF1TO10, LIST_OFF1TO10[val]))
        basic.append(rs)

        if _mem.settings.abr > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.abr
        rs = RadioSetting("settings.abr", "Backlight Timeout",
                          RadioSettingValueList(
                              LIST_OFF1TO10, LIST_OFF1TO10[val]))
        basic.append(rs)

        rs = RadioSetting("settings.tdr", "Dual Watch",
                          RadioSettingValueBoolean(_mem.settings.tdr))
        basic.append(rs)

        rs = RadioSetting("settings.beep", "Beep",
                          RadioSettingValueBoolean(_mem.settings.beep))
        basic.append(rs)

        if _mem.settings.timeout > 0x27:
            val = 0x03
        else:
            val = _mem.settings.timeout
        rs = RadioSetting("settings.timeout", "Timeout Timer",
                          RadioSettingValueList(
                              LIST_TIMEOUT, LIST_TIMEOUT[val]))
        basic.append(rs)

        if _mem.settings.voice > 0x02:
            val = 0x01
        else:
            val = _mem.settings.voice
        rs = RadioSetting("settings.voice", "Voice Prompt",
                          RadioSettingValueList(
                              LIST_VOICE, LIST_VOICE[val]))
        basic.append(rs)

        rs = RadioSetting("settings.dtmfst", "DTMF Sidetone",
                          RadioSettingValueList(LIST_DTMFST, LIST_DTMFST[
                              _mem.settings.dtmfst]))
        basic.append(rs)

        if _mem.settings.screv > 0x02:
            val = 0x01
        else:
            val = _mem.settings.screv
        rs = RadioSetting("settings.screv", "Scan Resume",
                          RadioSettingValueList(
                              LIST_RESUME, LIST_RESUME[val]))
        basic.append(rs)

        rs = RadioSetting("settings.pttid", "When to send PTT ID",
                          RadioSettingValueList(LIST_PTTID, LIST_PTTID[
                              _mem.settings.pttid]))
        basic.append(rs)

        if _mem.settings.pttlt > 0x1E:
            val = 0x05
        else:
            val = _mem.settings.pttlt
        rs = RadioSetting("pttlt", "PTT ID Delay",
                          RadioSettingValueInteger(0, 50, val))
        basic.append(rs)

        rs = RadioSetting("settings.mdfa", "Display Mode (A)",
                          RadioSettingValueList(LIST_MODE, LIST_MODE[
                              _mem.settings.mdfa]))
        basic.append(rs)

        rs = RadioSetting("settings.mdfb", "Display Mode (B)",
                          RadioSettingValueList(LIST_MODE, LIST_MODE[
                              _mem.settings.mdfb]))
        basic.append(rs)

        rs = RadioSetting("settings.autolk", "Automatic Key Lock",
                          RadioSettingValueBoolean(_mem.settings.autolk))
        basic.append(rs)

        rs = RadioSetting("settings.wtled", "Standby LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.wtled]))
        basic.append(rs)

        rs = RadioSetting("settings.rxled", "RX LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.rxled]))
        basic.append(rs)

        rs = RadioSetting("settings.txled", "TX LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.txled]))
        basic.append(rs)

        val = _mem.settings.almod
        rs = RadioSetting("settings.almod", "Alarm Mode",
                          RadioSettingValueList(
                              LIST_ALMOD, LIST_ALMOD[val]))
        basic.append(rs)

        if _mem.settings.tdrab > 0x02:
            val = 0x00
        else:
            val = _mem.settings.tdrab
        rs = RadioSetting("settings.tdrab", "Dual Watch TX Priority",
                          RadioSettingValueList(
                              LIST_OFFAB, LIST_OFFAB[val]))
        basic.append(rs)

        rs = RadioSetting("settings.ste", "Squelch Tail Eliminate (HT to HT)",
                          RadioSettingValueBoolean(_mem.settings.ste))
        basic.append(rs)

        if _mem.settings.rpste > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.rpste
        rs = RadioSetting("settings.rpste",
                          "Squelch Tail Eliminate (repeater)",
                          RadioSettingValueList(
                              LIST_RPSTE, LIST_RPSTE[val]))
        basic.append(rs)

        if _mem.settings.rptrl > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.rptrl
        rs = RadioSetting("settings.rptrl", "STE Repeater Delay",
                          RadioSettingValueList(
                              LIST_STEDELAY, LIST_STEDELAY[val]))
        basic.append(rs)

        rs = RadioSetting("settings.ponmsg", "Power-On Message",
                          RadioSettingValueList(LIST_PONMSG, LIST_PONMSG[
                              _mem.settings.ponmsg]))
        basic.append(rs)

        rs = RadioSetting("settings.roger", "Roger Beep",
                          RadioSettingValueBoolean(_mem.settings.roger))
        basic.append(rs)

        # Advanced settings
        rs = RadioSetting("settings.reset", "RESET Menu",
                          RadioSettingValueBoolean(_mem.settings.reset))
        advanced.append(rs)

        rs = RadioSetting("settings.menu", "All Menus",
                          RadioSettingValueBoolean(_mem.settings.menu))
        advanced.append(rs)

        rs = RadioSetting("settings.fmradio", "Broadcast FM Radio",
                          RadioSettingValueBoolean(_mem.settings.fmradio))
        advanced.append(rs)

        rs = RadioSetting("settings.alarm", "Alarm Sound",
                          RadioSettingValueBoolean(_mem.settings.alarm))
        advanced.append(rs)

        # Other settings
        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        _msg = _mem.firmware_msg
        val = RadioSettingValueString(0, 7, _filter(_msg.line1))
        val.set_mutable(False)
        rs = RadioSetting("firmware_msg.line1", "Firmware Message 1", val)
        other.append(rs)

        val = RadioSettingValueString(0, 7, _filter(_msg.line2))
        val.set_mutable(False)
        rs = RadioSetting("firmware_msg.line2", "Firmware Message 2", val)
        other.append(rs)

        _msg = _mem.sixpoweron_msg
        val = RadioSettingValueString(0, 7, _filter(_msg.line1))
        val.set_mutable(False)
        rs = RadioSetting("sixpoweron_msg.line1", "6+Power-On Message 1", val)
        other.append(rs)
        val = RadioSettingValueString(0, 7, _filter(_msg.line2))
        val.set_mutable(False)
        rs = RadioSetting("sixpoweron_msg.line2", "6+Power-On Message 2", val)
        other.append(rs)

        _msg = _mem.poweron_msg
        rs = RadioSetting("poweron_msg.line1", "Power-On Message 1",
                          RadioSettingValueString(
                              0, 7, _filter(_msg.line1)))
        other.append(rs)
        rs = RadioSetting("poweron_msg.line2", "Power-On Message 2",
                          RadioSettingValueString(
                              0, 7, _filter(_msg.line2)))
        other.append(rs)

        lower = 130
        upper = 179
        rs = RadioSetting("limits.vhf.lower", "VHF Lower Limit (MHz)",
                          RadioSettingValueInteger(
                              lower, upper, _mem.limits.vhf.lower))
        other.append(rs)

        rs = RadioSetting("limits.vhf.upper", "VHF Upper Limit (MHz)",
                          RadioSettingValueInteger(
                              lower, upper, _mem.limits.vhf.upper))
        other.append(rs)

        if self._tri_band:
            lower = 200
            upper = 260
            rs = RadioSetting("limits.vhf2.lower", "VHF2 Lower Limit (MHz)",
                              RadioSettingValueInteger(
                                  lower, upper, _mem.limits.vhf2.lower))
            other.append(rs)

            rs = RadioSetting("limits.vhf2.upper", "VHF2 Upper Limit (MHz)",
                              RadioSettingValueInteger(
                                  lower, upper, _mem.limits.vhf2.upper))
            other.append(rs)

        lower = 400
        upper = 520
        rs = RadioSetting("limits.uhf.lower", "UHF Lower Limit (MHz)",
                          RadioSettingValueInteger(
                              lower, upper, _mem.limits.uhf.lower))
        other.append(rs)

        rs = RadioSetting("limits.uhf.upper", "UHF Upper Limit (MHz)",
                          RadioSettingValueInteger(
                              lower, upper, _mem.limits.uhf.upper))
        other.append(rs)

        # Work mode settings
        rs = RadioSetting("settings.displayab", "Display",
                          RadioSettingValueList(
                              LIST_AB, LIST_AB[_mem.settings.displayab]))
        work.append(rs)

        rs = RadioSetting("settings.workmode", "VFO/MR Mode",
                          RadioSettingValueList(
                              LIST_WORKMODE,
                              LIST_WORKMODE[_mem.settings.workmode]))
        work.append(rs)

        rs = RadioSetting("settings.keylock", "Keypad Lock",
                          RadioSettingValueBoolean(_mem.settings.keylock))
        work.append(rs)

        rs = RadioSetting("wmchannel.mrcha", "MR A Channel",
                          RadioSettingValueInteger(0, 127,
                                                   _mem.wmchannel.mrcha))
        work.append(rs)

        rs = RadioSetting("wmchannel.mrchb", "MR B Channel",
                          RadioSettingValueInteger(0, 127,
                                                   _mem.wmchannel.mrchb))
        work.append(rs)

        def my_validate(value):
            value = chirp_common.parse_freq(value)
            msg = ("Can't be less than %i.0000")
            if value > 99000000 and value < 130 * 1000000:
                raise InvalidValueError(msg % (130))
            msg = ("Can't be between %i.9975-%i.0000")
            if (179 + 1) * 1000000 <= value and value < 400 * 1000000:
                raise InvalidValueError(msg % (179, 400))
            msg = ("Can't be greater than %i.9975")
            if value > 99000000 and value > (520 + 1) * 1000000:
                raise InvalidValueError(msg % (520))
            return chirp_common.format_freq(value)

        def apply_freq(setting, obj):
            value = chirp_common.parse_freq(str(setting.value)) / 10
            for i in range(7, -1, -1):
                obj.freq[i] = value % 10
                value /= 10

        val1a = RadioSettingValueString(0, 10,
                                        bfc.bcd_decode_freq(_mem.vfo.a.freq))
        val1a.set_validate_callback(my_validate)
        rs = RadioSetting("vfo.a.freq", "VFO A Frequency", val1a)
        rs.set_apply_callback(apply_freq, _mem.vfo.a)
        work.append(rs)

        val1b = RadioSettingValueString(0, 10,
                                        bfc.bcd_decode_freq(_mem.vfo.b.freq))
        val1b.set_validate_callback(my_validate)
        rs = RadioSetting("vfo.b.freq", "VFO B Frequency", val1b)
        rs.set_apply_callback(apply_freq, _mem.vfo.b)
        work.append(rs)

        rs = RadioSetting("vfo.a.sftd", "VFO A Shift",
                          RadioSettingValueList(
                              LIST_SHIFTD, LIST_SHIFTD[_mem.vfo.a.sftd]))
        work.append(rs)

        rs = RadioSetting("vfo.b.sftd", "VFO B Shift",
                          RadioSettingValueList(
                              LIST_SHIFTD, LIST_SHIFTD[_mem.vfo.b.sftd]))
        work.append(rs)

        def convert_bytes_to_offset(bytes):
            real_offset = 0
            for byte in bytes:
                real_offset = (real_offset * 10) + byte
            return chirp_common.format_freq(real_offset * 1000)

        def apply_offset(setting, obj):
            value = chirp_common.parse_freq(str(setting.value)) / 1000
            for i in range(5, -1, -1):
                obj.offset[i] = value % 10
                value /= 10

        val1a = RadioSettingValueString(
                    0, 10, convert_bytes_to_offset(_mem.vfo.a.offset))
        rs = RadioSetting("vfo.a.offset",
                          "VFO A Offset", val1a)
        rs.set_apply_callback(apply_offset, _mem.vfo.a)
        work.append(rs)

        val1b = RadioSettingValueString(
                    0, 10, convert_bytes_to_offset(_mem.vfo.b.offset))
        rs = RadioSetting("vfo.b.offset",
                          "VFO B Offset", val1b)
        rs.set_apply_callback(apply_offset, _mem.vfo.b)
        work.append(rs)

        def apply_txpower_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(
                      setting.value) + " from list")
            val = str(setting.value)
            index = TXP_CHOICES.index(val)
            val = TXP_VALUES[index]
            obj.set_value(val)

        if self._tri_band:
            if _mem.vfo.a.txpower3 in TXP_VALUES:
                idx = TXP_VALUES.index(_mem.vfo.a.txpower3)
            else:
                idx = TXP_VALUES.index(0x00)
            rs = RadioSettingValueList(TXP_CHOICES, TXP_CHOICES[idx])
            rset = RadioSetting("vfo.a.txpower3", "VFO A Power", rs)
            rset.set_apply_callback(apply_txpower_listvalue,
                                    _mem.vfo.a.txpower3)
            work.append(rset)

            if _mem.vfo.b.txpower3 in TXP_VALUES:
                idx = TXP_VALUES.index(_mem.vfo.b.txpower3)
            else:
                idx = TXP_VALUES.index(0x00)
            rs = RadioSettingValueList(TXP_CHOICES, TXP_CHOICES[idx])
            rset = RadioSetting("vfo.b.txpower3", "VFO B Power", rs)
            rset.set_apply_callback(apply_txpower_listvalue,
                                    _mem.vfo.b.txpower3)
            work.append(rset)
        else:
            rs = RadioSetting("vfo.a.txpower3", "VFO A Power",
                              RadioSettingValueList(
                                  LIST_TXPOWER,
                                  LIST_TXPOWER[min(_mem.vfo.a.txpower3, 0x02)]
                                               ))
            work.append(rs)

            rs = RadioSetting("vfo.b.txpower3", "VFO B Power",
                              RadioSettingValueList(
                                  LIST_TXPOWER,
                                  LIST_TXPOWER[min(_mem.vfo.b.txpower3, 0x02)]
                                               ))
            work.append(rs)

        rs = RadioSetting("vfo.a.widenarr", "VFO A Bandwidth",
                          RadioSettingValueList(
                              LIST_BANDWIDTH,
                              LIST_BANDWIDTH[_mem.vfo.a.widenarr]))
        work.append(rs)

        rs = RadioSetting("vfo.b.widenarr", "VFO B Bandwidth",
                          RadioSettingValueList(
                              LIST_BANDWIDTH,
                              LIST_BANDWIDTH[_mem.vfo.b.widenarr]))
        work.append(rs)

        rs = RadioSetting("vfo.a.scode", "VFO A S-CODE",
                          RadioSettingValueList(
                              LIST_SCODE,
                              LIST_SCODE[_mem.vfo.a.scode]))
        work.append(rs)

        rs = RadioSetting("vfo.b.scode", "VFO B S-CODE",
                          RadioSettingValueList(
                              LIST_SCODE,
                              LIST_SCODE[_mem.vfo.b.scode]))
        work.append(rs)

        rs = RadioSetting("vfo.a.step", "VFO A Tuning Step",
                          RadioSettingValueList(
                              LIST_STEP, LIST_STEP[_mem.vfo.a.step]))
        work.append(rs)
        rs = RadioSetting("vfo.b.step", "VFO B Tuning Step",
                          RadioSettingValueList(
                              LIST_STEP, LIST_STEP[_mem.vfo.b.step]))
        work.append(rs)

        # broadcast FM settings
        value = self._memobj.fm_presets
        value_shifted = ((value & 0x00FF) << 8) | ((value & 0xFF00) >> 8)
        if value_shifted >= 65.0 * 10 and value_shifted <= 108.0 * 10:
            # storage method 3 (discovered 2022)
            self._bw_shift = True
            preset = value_shifted / 10.0
        elif value >= 65.0 * 10 and value <= 108.0 * 10:
            # storage method 2
            preset = value / 10.0
        elif value <= 108.0 * 10 - 650:
            # original storage method (2012)
            preset = value / 10.0 + 65
        else:
            # unknown (undiscovered method or no FM chip?)
            preset = False
        if preset:
            rs = RadioSettingValueFloat(65, 108.0, preset, 0.1, 1)
            rset = RadioSetting("fm_presets", "FM Preset(MHz)", rs)
            fm_preset.append(rset)

        # DTMF settings
        def apply_code(setting, obj, length):
            code = []
            for j in range(0, length):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.code = code

        for i in range(0, 15):
            _codeobj = self._memobj.pttid[i].code
            _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            val = RadioSettingValueString(0, 5, _code, False)
            val.set_charset(DTMF_CHARS)
            pttid = RadioSetting("pttid/%i.code" % i,
                                 "Signal Code %i" % (i + 1), val)
            pttid.set_apply_callback(apply_code, self._memobj.pttid[i], 5)
            dtmfe.append(pttid)

        if _mem.ani.dtmfon > 0xC3:
            val = 0x03
        else:
            val = _mem.ani.dtmfon
        rs = RadioSetting("ani.dtmfon", "DTMF Speed (on)",
                          RadioSettingValueList(LIST_DTMFSPEED,
                                                LIST_DTMFSPEED[val]))
        dtmfe.append(rs)

        if _mem.ani.dtmfoff > 0xC3:
            val = 0x03
        else:
            val = _mem.ani.dtmfoff
        rs = RadioSetting("ani.dtmfoff", "DTMF Speed (off)",
                          RadioSettingValueList(LIST_DTMFSPEED,
                                                LIST_DTMFSPEED[val]))
        dtmfe.append(rs)

        _codeobj = self._memobj.ani.code
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 5, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("ani.code", "ANI Code", val)
        rs.set_apply_callback(apply_code, self._memobj.ani, 5)
        dtmfe.append(rs)

        rs = RadioSetting("ani.aniid", "When to send ANI ID",
                          RadioSettingValueList(LIST_PTTID,
                                                LIST_PTTID[_mem.ani.aniid]))
        dtmfe.append(rs)

        # Service settings
        for band in ["vhf", "uhf"]:
            for index in range(0, 10):
                key = "squelch.%s.sql%i" % (band, index)
                if band == "vhf":
                    _obj = self._memobj.squelch.vhf
                elif band == "uhf":
                    _obj = self._memobj.squelch.uhf
                val = RadioSettingValueInteger(0, 123,
                                               getattr(
                                                   _obj, "sql%i" % (index)))
                if index == 0:
                    val.set_mutable(False)
                name = "%s Squelch %i" % (band.upper(), index)
                rs = RadioSetting(key, name, val)
                service.append(rs)

        return top

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) in [0x2008, 0x2010]:
            match_size = True

        # testing the firmware model fingerprint
        match_model = model_match(cls, filedata)

        if match_size and match_model:
            return True
        else:
            return False


class RH5XAlias(chirp_common.Alias):
    VENDOR = "Rugged"
    MODEL = "RH5X"


class UV82IIIAlias(chirp_common.Alias):
    VENDOR = "Baofeng"
    MODEL = "UV-82III"


class UV9RPROAlias(chirp_common.Alias):
    VENDOR = "Baofeng"
    MODEL = "UV-9R Pro"


@directory.register
class BFA58(WP970I):
    """Baofeng BF-A58"""
    VENDOR = "Baofeng"
    MODEL = "BF-A58"
    LENGTH_NAME = 7
    ALIASES = [RH5XAlias, UV9RPROAlias]

    _fileid = ["BFT515 ", "BFT517 "]


@directory.register
class UV82WP(WP970I):
    """Baofeng UV82-WP"""
    VENDOR = "Baofeng"
    MODEL = "UV-82WP"


@directory.register
class GT3WP(WP970I):
    """Baofeng GT-3WP"""
    VENDOR = "Baofeng"
    MODEL = "GT-3WP"
    LENGTH_NAME = 7


@directory.register
class RT6(WP970I):
    """Retevis RT6"""
    VENDOR = "Retevis"
    MODEL = "RT6"


@directory.register
class BFA58S(WP970I):
    VENDOR = "Baofeng"
    MODEL = "BF-A58S"
    LENGTH_NAME = 7
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Low", watts=1.00)]
    ALIASES = [UV82IIIAlias]
    _tri_band = True

    def get_features(self):
        rf = WP970I.get_features(self)
        rf.valid_bands = [self._vhf_range,
                          self._vhf2_range,
                          self._uhf_range]
        return rf


@directory.register
class UVS9X3(BFA58S):
    VENDOR = "Baofeng"
    MODEL = "UV-S9X3"
    ALIASES = []
    _magic = [MSTRING_UVS9X3, ]


@directory.register
class BF5RXRadio(BFA58S):
    VENDOR = "Baofeng"
    MODEL = "5RX"
    ALIASES = []
    _magic = [MSTRING_UVS9X3, ]

    _air_range = (108000000, 136000000)
    _vhf_range = (136000000, 174000000)
    _vhf2_range = (200000000, 260000000)
    _uhf2_range = (350000000, 390000000)
    _uhf_range = (400000000, 520000000)

    def get_features(self):
        rf = WP970I.get_features(self)
        rf.valid_bands = [self._air_range,
                          self._vhf_range,
                          self._vhf2_range,
                          self._uhf2_range,
                          self._uhf_range]
        return rf


@directory.register
class UV9R(WP970I):
    """Baofeng UV-9R"""
    VENDOR = "Baofeng"
    MODEL = "UV-9R"
    LENGTH_NAME = 7


@directory.register
class UV9G(WP970I):
    """Baofeng UV-9G"""
    VENDOR = "Baofeng"
    MODEL = "UV-9G"
    LENGTH_NAME = 7

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Med",  watts=1.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]
    _magic = [MSTRING_UV9G, ]
    _gmrs = False  # sold as GMRS radio but supports full band TX/RX

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging
import re

from chirp.drivers import yaesu_clone
from chirp import chirp_common, memmap, directory, bitwise, errors
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueBoolean, \
    RadioSettingValueString, RadioSettings

from collections import defaultdict

LOG = logging.getLogger(__name__)

ACK = b'\x06'

MEM_FORMAT = """
struct mem_struct {
  u8 used:1,
     unknown1:1,
     mode:2,
     unknown2:1,
     duplex:3;
  bbcd freq[3];
  u8 clockshift:1,
     tune_step:3,
     unknown5:1, // TODO: tmode has extended settings, at least 4 bits
     tmode:3;
  bbcd split[3];
  u8 power:2,
     tone:6;
  u8 unknown6:1,
     dtcs:7;
  u8 unknown7[2];
  u8 offset;
  u8 unknown9[3];
};

#seekto 0x002A;
u8  banks_unk2;
u8  current_channel;
u8  unk3;
u8  unk4;
u8  current_menu;

#seekto 0x003A;
struct {
    u8  apo;
    u8  tot;
    u8  lock:3,
        arts_interval:1,
        unk1a:1,
        prog_panel_acc:3;
    u8  prog_p1;
    u8  prog_p2;
    u8  prog_p3;
    u8  prog_p4;
    u8  rf_sql;
    u8  inet_dtmf_mem:4,
        inet_dtmf_digit:4;
    u8  arts_cwid_enable:1,
        prog_tone_vm:1,
        unk2a:1,
        hyper_write:2,
        memory_only:1,
        dimmer:2;
    u8  beep_scan:1,
        beep_edge:1,
        beep_key:1,
        unk3a:1,
        inet_mode:1,
        unk3b:1,
        dtmf_speed:2;
    u8  dcs_polarity:2,
        smart_search:1,
        priority_revert:1,
        unk4a:1,
        dtmf_delay:3;
    u8  unk5a:3,
        microphone_type:1,
        scan_resume:1,
        unk5b:1,
        arts_mode:2;
    u8  unk6;
} settings;

#seekto 0x0048;
struct mem_struct vfos[5];

#seekto 0x00C8;
struct {
    u8  memory[16];
} dtmf[16];

#seekto 0x01C8;
struct mem_struct homes[5];

#seekto 0x0218;
u8  arts_cwid[6];

#seekto 0x04C8;
struct mem_struct memory[1000];

#seekto 0x4988;
struct {
  char name[6];
  u8 enabled:1,
     unknown1:7;
  u8 used:1,
     unknown2:7;
} names[1000];

#seekto 0x6c48;
struct {
   u32 bitmap[32];
} bank_channels[20];

#seekto 0x7648;
struct {
  u8 skip0:2,
     skip1:2,
     skip2:2,
     skip3:2;
} flags[250];

#seekto 0x7B48;
u8 checksum;
"""

MODES = ["FM", "AM", "NFM"]
DUPLEX = ["", "", "-", "+", "split"]
STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]
SKIPS = ["", "S", "P", ""]

CHARSET = ["%i" % int(x) for x in range(0, 10)] + \
    [chr(x) for x in range(ord("A"), ord("Z")+1)] + \
    list(" " * 10) + \
    list("*+,- /|      [ ] _") + \
    list("\x00" * 100)

DTMFCHARSET = list("0123456789ABCD*#")


def _send(ser, data):
    for i in data:
        ser.write(bytes([i]))
        time.sleep(0.002)
    echo = ser.read(len(data))
    if echo != data:
        raise errors.RadioError("Error reading echo (Bad cable?)")


def _download(radio):
    data = bytes(b"")

    chunk = bytes(b"")
    for i in range(0, 30):
        chunk += radio.pipe.read(radio._block_lengths[0])
        if chunk:
            break

    if len(chunk) != radio._block_lengths[0]:
        raise Exception("Failed to read header (%i)" % len(chunk))
    data += chunk

    _send(radio.pipe, ACK)

    for i in range(0, radio._block_lengths[1], radio._block_size):
        chunk = radio.pipe.read(radio._block_size)
        data += chunk
        if len(chunk) != radio._block_size:
            break
        time.sleep(0.01)
        _send(radio.pipe, ACK)
        if radio.status_fn:
            status = chirp_common.Status()
            status.max = radio.get_memsize()
            status.cur = i+len(chunk)
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    data += radio.pipe.read(1)
    _send(radio.pipe, ACK)

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    cur = 0
    mmap = radio.get_mmap().get_byte_compatible()
    for block in radio._block_lengths:
        for _i in range(0, block, radio._block_size):
            length = min(radio._block_size, block)
            # LOG.debug("i=%i length=%i range: %i-%i" %
            #           (i, length, cur, cur+length))
            _send(radio.pipe, mmap[cur:cur+length])
            if radio.pipe.read(1) != ACK:
                raise errors.RadioError("Radio did not ack block at %i" % cur)
            cur += length
            time.sleep(0.05)

            if radio.status_fn:
                status = chirp_common.Status()
                status.cur = cur
                status.max = radio.get_memsize()
                status.msg = "Cloning to radio"
                radio.status_fn(status)


def get_freq(rawfreq):
    """Decode a frequency that may include a fractional step flag"""
    # Ugh.  The 0x80 and 0x40 indicate values to add to get the
    # real frequency.  Gross.
    if rawfreq > 8000000000:
        rawfreq = (rawfreq - 8000000000) + 5000

    if rawfreq > 4000000000:
        rawfreq = (rawfreq - 4000000000) + 2500

    return rawfreq


def set_freq(freq, obj, field):
    """Encode a frequency with any necessary fractional step flags"""
    obj[field] = freq / 10000
    if (freq % 1000) == 500:
        obj[field][0].set_bits(0x40)

    if (freq % 10000) >= 5000:
        obj[field][0].set_bits(0x80)

    return freq


class FTx800Radio(yaesu_clone.YaesuCloneModeRadio):
    """Base class for FT-7800,7900,8800,8900 radios"""
    BAUD_RATE = 9600
    VENDOR = "Yaesu"
    MODES = list(MODES)
    NEEDS_COMPAT_SERIAL = False
    _block_size = 64

    POWER_LEVELS_VHF = [chirp_common.PowerLevel("Hi", watts=50),
                        chirp_common.PowerLevel("Mid1", watts=20),
                        chirp_common.PowerLevel("Mid2", watts=10),
                        chirp_common.PowerLevel("Low", watts=5)]

    POWER_LEVELS_UHF = [chirp_common.PowerLevel("Hi", watts=35),
                        chirp_common.PowerLevel("Mid1", watts=20),
                        chirp_common.PowerLevel("Mid2", watts=10),
                        chirp_common.PowerLevel("Low", watts=5)]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA jack.\n"
            "3. Press and hold in the [MHz(PRI)] key while turning the\n"
            " radio on.\n"
            "4. Rotate the DIAL job to select \"F-7 CLONE\".\n"
            "5. Press and hold in the [BAND(SET)] key. The display\n"
            " will disappear for a moment, then the \"CLONE\" notation\n"
            " will appear.\n"
            "6. <b>After clicking OK</b>, press the [V/M(MW)] key to send"
            " image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA jack.\n"
            "3. Press and hold in the [MHz(PRI)] key while turning the\n"
            "     radio on.\n"
            "4. Rotate the DIAL job to select \"F-7 CLONE\".\n"
            "5. Press and hold in the [BAND(SET)] key. The display\n"
            "     will disappear for a moment, then the \"CLONE\" notation\n"
            "     will appear.\n"
            "6. Press the [LOW(ACC)] key (\"--RX--\" will appear on the"
            " display).\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 999)
        rf.has_bank = False
        rf.has_ctone = False
        rf.has_dtcs_polarity = False
        rf.valid_modes = MODES
        rf.valid_tmodes = self.TMODES
        rf.valid_duplexes = ["", "-", "+", "split"]
        rf.valid_tuning_steps = STEPS
        rf.valid_bands = [(108000000, 520000000), (700000000, 990000000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_power_levels = self.POWER_LEVELS_VHF
        rf.valid_characters = "".join(CHARSET)
        rf.valid_name_length = 6
        rf.can_odd_split = True
        return rf

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0x7B47)]

    def sync_in(self):
        start = time.time()
        try:
            self._mmap = _download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Download finished in %i seconds" % (time.time() - start))
        self.check_checksums()
        self.process_mmap()

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_out(self):
        self.update_checksums()
        start = time.time()
        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Upload finished in %i seconds" % (time.time() - start))

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number-1])

    def _get_mem_offset(self, mem, _mem):
        if mem.duplex == "split":
            return get_freq(int(_mem.split) * 10000)
        else:
            return (_mem.offset * 5) * 10000

    def _set_mem_offset(self, mem, _mem):
        if mem.duplex == "split":
            set_freq(mem.offset, _mem, "split")
        else:
            _mem.offset = (int(mem.offset / 10000) / 5)

    def _get_mem_name(self, mem, _mem):
        _nam = self._memobj.names[mem.number - 1]

        name = ""
        if _nam.used:
            for i in str(_nam.name):
                name += CHARSET[ord(i)]

        return name.rstrip()

    def _set_mem_name(self, mem, _mem):
        _nam = self._memobj.names[mem.number - 1]

        if mem.name.rstrip():
            name = [chr(CHARSET.index(x)) for x in mem.name.ljust(6)[:6]]
            _nam.name = "".join(name)
            _nam.used = 1
            _nam.enabled = 1
        else:
            _nam.used = 0
            _nam.enabled = 0

    def _get_mem_skip(self, mem, _mem):
        _flg = self._memobj.flags[(mem.number - 1) / 4]
        flgidx = (mem.number - 1) % 4
        return SKIPS[_flg["skip%i" % flgidx]]

    def _set_mem_skip(self, mem, _mem):
        _flg = self._memobj.flags[(mem.number - 1) / 4]
        flgidx = (mem.number - 1) % 4
        _flg["skip%i" % flgidx] = SKIPS.index(mem.skip)

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()
        mem.number = number
        mem.empty = not _mem.used
        if mem.empty:
            return mem

        mem.freq = get_freq(int(_mem.freq) * 10000)
        mem.rtone = chirp_common.TONES[_mem.tone]
        mem.tmode = self.TMODES[_mem.tmode]
        mem.mode = self.MODES[_mem.mode]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        if self.get_features().has_tuning_step:
            mem.tuning_step = STEPS[_mem.tune_step]
        mem.duplex = DUPLEX[_mem.duplex]
        mem.offset = self._get_mem_offset(mem, _mem)
        mem.name = self._get_mem_name(mem, _mem)

        if int(mem.freq / 100) == 4:
            mem.power = self.POWER_LEVELS_UHF[_mem.power]
        else:
            mem.power = self.POWER_LEVELS_VHF[_mem.power]

        mem.skip = self._get_mem_skip(mem, _mem)

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        _mem.used = int(not mem.empty)
        if mem.empty:
            return

        set_freq(mem.freq, _mem, "freq")
        _mem.tone = chirp_common.TONES.index(mem.rtone)
        _mem.tmode = self.TMODES.index(mem.tmode)
        _mem.mode = self.MODES.index(mem.mode)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        if self.get_features().has_tuning_step:
            _mem.tune_step = STEPS.index(mem.tuning_step)
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.split = mem.duplex == "split" and int(mem.offset / 10000) or 0
        if mem.power:
            _mem.power = self.POWER_LEVELS_VHF.index(mem.power)
        else:
            _mem.power = 0
        _mem.unknown5 = 0  # Make sure we don't leave garbage here

        # NB: Leave offset after mem name for the 8800!
        self._set_mem_name(mem, _mem)
        self._set_mem_offset(mem, _mem)

        self._set_mem_skip(mem, _mem)


class FT7800BankModel(chirp_common.BankModel):
    """Yaesu FT-7800/7900 bank model"""
    def __init__(self, radio):
        super(FT7800BankModel, self).__init__(radio)
        self.__b2m_cache = defaultdict(list)
        self.__m2b_cache = defaultdict(list)

    def __precache(self):
        if self.__b2m_cache:
            return

        for bank in self.get_mappings():
            self.__b2m_cache[bank.index] = self._get_bank_memories(bank)
            for memnum in self.__b2m_cache[bank.index]:
                self.__m2b_cache[memnum].append(bank.index)

    def get_num_mappings(self):
        return 20

    def get_mappings(self):
        banks = []
        for i in range(0, self.get_num_mappings()):
            bank = chirp_common.Bank(self, "%i" % i, "BANK-%i" % (i + 1))
            bank.index = i
            banks.append(bank)

        return banks

    def add_memory_to_mapping(self, memory, bank):
        self.__precache()

        index = memory.number - 1
        _bitmap = self._radio._memobj.bank_channels[bank.index]
        ishft = 31 - (index % 32)
        _bitmap.bitmap[index // 32] |= (1 << ishft)
        self.__m2b_cache[memory.number].append(bank.index)
        self.__b2m_cache[bank.index].append(memory.number)

    def remove_memory_from_mapping(self, memory, bank):
        self.__precache()

        index = memory.number - 1
        _bitmap = self._radio._memobj.bank_channels[bank.index]
        ishft = 31 - (index % 32)
        if not (_bitmap.bitmap[index // 32] & (1 << ishft)):
            raise Exception("Memory {num} is not in bank {bank}".format(
                            num=memory.number, bank=bank))
        _bitmap.bitmap[index // 32] &= ~(1 << ishft)
        self.__b2m_cache[bank.index].remove(memory.number)
        self.__m2b_cache[memory.number].remove(bank.index)

    def _get_bank_memories(self, bank):
        memories = []
        upper = self._radio.get_features().memory_bounds[1]
        c = self._radio._memobj.bank_channels[bank.index]
        for i in range(0, upper):
            _bitmap = c.bitmap[i // 32]
            ishft = 31 - (i % 32)
            if _bitmap & (1 << ishft):
                memories.append(i + 1)
        return memories

    def get_mapping_memories(self, bank):
        self.__precache()

        return [self._radio.get_memory(n)
                for n in self.__b2m_cache[bank.index]]

    def get_memory_mappings(self, memory):
        self.__precache()

        _banks = self.get_mappings()
        return [_banks[b] for b in self.__m2b_cache[memory.number]]


@directory.register
class FT7800Radio(FTx800Radio):
    """Yaesu FT-7800"""
    MODEL = "FT-7800/7900"

    _model = b"AH016"
    _memsize = 31561
    _block_lengths = [8, 31552, 1]
    TMODES = ["", "Tone", "TSQL", "TSQL-R", "DTCS"]

    def get_bank_model(self):
        if not hasattr(self, '_banks'):
            self._banks = FT7800BankModel(self)
        return self._banks

    def get_features(self):
        rf = FTx800Radio.get_features(self)
        rf.has_bank = True
        rf.has_settings = True
        return rf

    def set_memory(self, memory):
        if memory.empty:
            self._wipe_memory_banks(memory)
        FTx800Radio.set_memory(self, memory)

    def _decode_chars(self, inarr):
        LOG.debug("@_decode_chars, type: %s" % type(inarr))
        LOG.debug(inarr)
        outstr = ""
        for i in inarr:
            if i == 0xFF:
                break
            outstr += CHARSET[i & 0x7F]
        return outstr.rstrip()

    def _encode_chars(self, instr, length=16):
        LOG.debug("@_encode_chars, type: %s" % type(instr))
        LOG.debug(instr)
        outarr = []
        instr = str(instr)
        for i in range(length):
            if i < len(instr):
                outarr.append(CHARSET.index(instr[i]))
            else:
                outarr.append(0xFF)
        return outarr

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic")
        dtmf = RadioSettingGroup("dtmf", "DTMF")
        arts = RadioSettingGroup("arts", "ARTS")
        prog = RadioSettingGroup("prog", "Programmable Buttons")

        top = RadioSettings(basic, dtmf, arts, prog)

        basic.append(RadioSetting(
                "priority_revert", "Priority Revert",
                RadioSettingValueBoolean(_settings.priority_revert)))

        basic.append(RadioSetting(
                "memory_only", "Memory Only mode",
                RadioSettingValueBoolean(_settings.memory_only)))

        opts = ["off"] + ["%0.1f" % (t / 60.0) for t in range(30, 750, 30)]
        basic.append(RadioSetting(
                "apo", "APO time (hrs)",
                RadioSettingValueList(opts, opts[_settings.apo])))

        basic.append(RadioSetting(
                "beep_scan", "Beep: Scan",
                RadioSettingValueBoolean(_settings.beep_scan)))

        basic.append(RadioSetting(
                "beep_edge", "Beep: Edge",
                RadioSettingValueBoolean(_settings.beep_edge)))

        basic.append(RadioSetting(
                "beep_key", "Beep: Key",
                RadioSettingValueBoolean(_settings.beep_key)))

        opts = ["T/RX Normal", "RX Reverse", "TX Reverse", "T/RX Reverse"]
        basic.append(RadioSetting(
                "dcs_polarity", "DCS polarity",
                RadioSettingValueList(opts, opts[_settings.dcs_polarity])))

        opts = ["off", "dim 1", "dim 2", "dim 3"]
        basic.append(RadioSetting(
                "dimmer", "Dimmer",
                RadioSettingValueList(opts, opts[_settings.dimmer])))

        opts = ["manual", "auto", "1-auto"]
        basic.append(RadioSetting(
                "hyper_write", "Hyper Write",
                RadioSettingValueList(opts, opts[_settings.hyper_write])))

        opts = ["", "key", "dial", "key+dial", "ptt",
                "ptt+key", "ptt+dial", "all"]
        basic.append(RadioSetting(
                "lock", "Lock mode",
                RadioSettingValueList(opts, opts[_settings.lock])))

        opts = ["MH-42", "MH-48"]
        basic.append(RadioSetting(
                "microphone_type", "Microphone Type",
                RadioSettingValueList(opts, opts[_settings.microphone_type])))

        opts = ["off"] + ["S-%d" % n for n in range(2, 10)] + ["S-Full"]
        basic.append(RadioSetting(
                "rf_sql", "RF Squelch",
                RadioSettingValueList(opts, opts[_settings.rf_sql])))

        opts = ["time", "hold", "busy"]
        basic.append(RadioSetting(
                "scan_resume", "Scan Resume",
                RadioSettingValueList(opts, opts[_settings.scan_resume])))

        opts = ["single", "continuous"]
        basic.append(RadioSetting(
                "smart_search", "Smart Search",
                RadioSettingValueList(opts, opts[_settings.smart_search])))

        opts = ["off"] + ["%d" % t for t in range(1, 31)]
        basic.append(RadioSetting(
                "tot", "Time-out timer (mins)",
                RadioSettingValueList(opts, opts[_settings.tot])))

        # dtmf tab

        opts = ["50", "100", "250", "450", "750", "1000"]
        dtmf.append(RadioSetting(
                "dtmf_delay", "DTMF delay (ms)",
                RadioSettingValueList(opts, opts[_settings.dtmf_delay])))

        opts = ["50", "75", "100"]
        dtmf.append(RadioSetting(
                "dtmf_speed", "DTMF speed (ms)",
                RadioSettingValueList(opts, opts[_settings.dtmf_speed])))

        for i in range(16):
            name = "dtmf%02d" % i
            dtmfsetting = self._memobj.dtmf[i]
            dtmfstr = ""
            for c in dtmfsetting.memory:
                if c == 0xFF:
                    break
                if c < len(DTMFCHARSET):
                    dtmfstr += DTMFCHARSET[c]
            LOG.debug(dtmfstr)
            dtmfentry = RadioSettingValueString(0, 16, dtmfstr)
            dtmfentry.set_charset(DTMFCHARSET + list(" "))
            rs = RadioSetting(name, name.upper(), dtmfentry)
            dtmf.append(rs)

        # arts tab

        opts = ["off", "in range", "always"]
        arts.append(RadioSetting(
                "arts_mode", "ARTS beep",
                RadioSettingValueList(opts, opts[_settings.arts_mode])))

        opts = ["15", "25"]
        arts.append(RadioSetting(
                "arts_interval", "ARTS interval",
                RadioSettingValueList(opts, opts[_settings.arts_interval])))

        arts.append(RadioSetting(
                "arts_cwid_enable", "CW ID",
                RadioSettingValueBoolean(_settings.arts_cwid_enable)))

        _arts_cwid = self._memobj.arts_cwid
        cwid = RadioSettingValueString(
                0, 16, self._decode_chars(_arts_cwid.get_value()))
        cwid.set_charset(CHARSET)
        arts.append(RadioSetting("arts_cwid", "CW ID", cwid))

        # prog buttons

        opts = ["WX", "Reverse", "Repeater", "SQL Off", "Lock", "Dimmer"]
        prog.append(RadioSetting(
                "prog_panel_acc", "Prog Panel - Low(ACC)",
                RadioSettingValueList(opts, opts[_settings.prog_panel_acc])))

        opts = ["Reverse", "Home"]
        prog.append(RadioSetting(
                "prog_tone_vm", "TONE | V/M",
                RadioSettingValueList(opts, opts[_settings.prog_tone_vm])))

        opts = ["" for n in range(26)] + \
            ["Priority", "Low", "Tone", "MHz", "Reverse", "Home", "Band",
             "VFO/MR", "Scan", "Sql Off", "TCall", "SSCH", "ARTS", "Tone Freq",
             "DCSC", "WX", "Repeater"]

        prog.append(RadioSetting(
                "prog_p1", "P1",
                RadioSettingValueList(opts, opts[_settings.prog_p1])))

        prog.append(RadioSetting(
                "prog_p2", "P2",
                RadioSettingValueList(opts, opts[_settings.prog_p2])))

        prog.append(RadioSetting(
                "prog_p3", "P3",
                RadioSettingValueList(opts, opts[_settings.prog_p3])))

        prog.append(RadioSetting(
                "prog_p4", "P4",
                RadioSettingValueList(opts, opts[_settings.prog_p4])))

        return top

    def set_settings(self, uisettings):
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                _settings = self._memobj.settings
                setting = element.get_name()
                if re.match(r'dtmf\d', setting):
                    # set dtmf fields
                    dtmfstr = str(element.value).strip()
                    newval = []
                    for i in range(0, 16):
                        if i < len(dtmfstr):
                            newval.append(DTMFCHARSET.index(dtmfstr[i]))
                        else:
                            newval.append(0xFF)
                    LOG.debug(newval)
                    idx = int(setting[-2:])
                    _settings = self._memobj.dtmf[idx]
                    _settings.memory = newval
                    continue
                if setting == "arts_cwid":
                    oldval = self._memobj.arts_cwid
                    newval = self._encode_chars(newval.get_value(), 6)
                    self._memobj.arts_cwid = newval
                    continue
                # normal settings
                newval = element.value
                oldval = getattr(_settings, setting)
                LOG.debug("Setting %s(%s) <= %s" % (setting, oldval, newval))
                setattr(_settings, setting, newval)
            except Exception:
                LOG.debug(element.get_name())
                raise


MEM_FORMAT_8800 = """
#seekto 0x%X;
struct {
  u8 used:1,
     unknown1:1,
     mode:2,
     unknown2:1,
     duplex:3;
  bbcd freq[3];
  u8 unknown3:1,
     tune_step:3,
     power:2,
     tmode:2;
  bbcd split[3];
  u8 nameused:1,
     unknown5:1,
     tone:6;
  u8 namevalid:1,
     dtcs:7;
  u8 name[6];
} memory[500];

#seekto 0x%X;
struct {
   u32 bitmap[16];
} bank_channels[10];

#seekto 0x51C8;
struct {
  u8 skip0:2,
     skip1:2,
     skip2:2,
     skip3:2;
} flags[250];

#seekto 0x7B48;
u8 checksum;
"""


class FT8800BankModel(FT7800BankModel):
    def get_num_mappings(self):
        return 10


@directory.register
class FT8800Radio(FTx800Radio):
    """Base class for Yaesu FT-8800"""
    MODEL = "FT-8800"

    _model = b"AH018"
    _memsize = 22217

    _block_lengths = [8, 22208, 1]

    _memstart = 0x0000

    TMODES = ["", "Tone", "TSQL", "DTCS"]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA jack.\n"
            "3. Press and hold in the \"left\" [V/M] key while turning the\n"
            "     radio on.\n"
            "4. Rotate the \"right\" DIAL knob to select \"CLONE START\".\n"
            "5. Press the [SET] key. The display will disappear\n"
            "     for a moment, then the \"CLONE\" notation will appear.\n"
            "6. <b>After clicking OK</b>, press the \"left\" [V/M] key to\n"
            "     send image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA jack.\n"
            "3. Press and hold in the \"left\" [V/M] key while turning the\n"
            "     radio on.\n"
            "4. Rotate the \"right\" DIAL knob to select \"CLONE START\".\n"
            "5. Press the [SET] key. The display will disappear\n"
            "     for a moment, then the \"CLONE\" notation will appear.\n"
            "6. Press the \"left\" [LOW] key (\"CLONE -RX-\" will appear"
            " on\n"
            "     the display).\n")
        return rp

    def get_features(self):
        rf = FTx800Radio.get_features(self)
        rf.has_sub_devices = self.VARIANT == ""
        rf.has_bank = True
        rf.memory_bounds = (1, 500)
        return rf

    def get_sub_devices(self):
        return [FT8800RadioLeft(self._mmap), FT8800RadioRight(self._mmap)]

    def get_bank_model(self):
        if not hasattr(self, '_banks'):
            self._banks = FT8800BankModel(self)
        return self._banks

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0x56C7)]

    def process_mmap(self):
        if not self._memstart:
            return

        self._memobj = bitwise.parse(MEM_FORMAT_8800 % (self._memstart,
                                                        self._bankstart),
                                     self._mmap)

    def _get_mem_offset(self, mem, _mem):
        if mem.duplex == "split":
            return get_freq(int(_mem.split) * 10000)

        # The offset is packed into the upper two bits of the last four
        # bytes of the name (?!)
        val = 0
        for i in _mem.name[2:6]:
            val <<= 2
            val |= ((i & 0xC0) >> 6)

        return (val * 5) * 10000

    def _set_mem_offset(self, mem, _mem):
        if mem.duplex == "split":
            set_freq(mem.offset, _mem, "split")
            return

        val = int(mem.offset / 10000) // 5
        for i in reversed(list(range(2, 6))):
            _mem.name[i] = (_mem.name[i] & 0x3F) | ((val & 0x03) << 6)
            val >>= 2

    def _get_mem_name(self, mem, _mem):
        name = ""
        if _mem.namevalid:
            for i in _mem.name:
                index = int(i) & 0x3F
                if index < len(CHARSET):
                    name += CHARSET[index]

        return name.rstrip()

    def _set_mem_name(self, mem, _mem):
        _mem.name = [CHARSET.index(x) for x in mem.name.ljust(6)[:6]]
        _mem.namevalid = 1
        _mem.nameused = bool(mem.name.rstrip())


class FT8800RadioLeft(FT8800Radio):
    """Yaesu FT-8800 Left VFO subdevice"""
    VARIANT = "Left"
    _memstart = 0x0948
    _bankstart = 0x4BC8


class FT8800RadioRight(FT8800Radio):
    """Yaesu FT-8800 Right VFO subdevice"""
    VARIANT = "Right"
    _memstart = 0x2948
    _bankstart = 0x4BC8


MEM_FORMAT_8900 = """
#seekto 0x0708;
struct {
  u8 used:1,
     skip:2,
     sub_used:1,
     unknown2:1,
     duplex:3;
  bbcd freq[3];
  u8 mode:2,
     nameused:1,
     unknown4:1,
     power:2,
     tmode:2;
  bbcd split[3];
  u8 unknown5:2,
     tone:6;
  u8 namevalid:1,
     dtcs:7;
  u8 name[6];
} memory[799];

#seekto 0x51C8;
struct {
  u8 skip0:2,
     skip1:2,
     skip2:2,
     skip3:2;
} flags[400];

#seekto 0x7B48;
u8 checksum;
"""


@directory.register
class FT8900Radio(FT8800Radio):
    """Yaesu FT-8900"""
    MODEL = "FT-8900"

    _model = b"AH008"
    _memsize = 14793
    _block_lengths = [8, 14784, 1]

    MODES = ["FM", "NFM", "AM"]

    def get_bank_model(self):
        return

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT_8900, self._mmap)

    def get_features(self):
        rf = FT8800Radio.get_features(self)
        rf.has_sub_devices = False
        rf.has_bank = False
        rf.valid_modes = self.MODES
        rf.valid_bands = [(28000000,  29700000),
                          (50000000,  54000000),
                          (108000000, 180000000),
                          (320000000, 480000000),
                          (700000000, 985000000)]
        rf.memory_bounds = (1, 799)
        rf.has_tuning_step = False

        return rf

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0x39C7)]

    def _get_mem_skip(self, mem, _mem):
        return SKIPS[_mem.skip]

    def _set_mem_skip(self, mem, _mem):
        _mem.skip = SKIPS.index(mem.skip)

    def get_memory(self, number):
        mem = super().get_memory(number)

        _mem = self._memobj.memory[number - 1]

        return mem

    def set_memory(self, mem):
        FT8800Radio.set_memory(self, mem)

        # The 8900 has a bit flag that tells the radio whether or not
        # the memory should show up on the sub (right) band
        _mem = self._memobj.memory[mem.number - 1]
        if mem.freq < 108000000 or mem.freq > 480000000:
            _mem.sub_used = 0
        else:
            _mem.sub_used = 1
